REM ---====================== Start Software License ======================---
REM --==                                                                  ==--
REM --== This license governs the use of this software, and your use of   ==--
REM --== this software constitutes acceptance of this license. Agreement  ==--
REM --== with all points is required to use this software.                ==--
REM --==                                                                  ==--
REM --== 1. You may use this software freely for personal use.            ==--
REM --==                                                                  ==--
REM --== 2. You may use this software freely to determine feasibility for ==--
REM --==    commercial use.                                               ==--
REM --==                                                                  ==--
REM --== 3. You may use this software for commercial use if the author    ==--
REM --==    has given you written consent.                                ==--
REM --==                                                                  ==--
REM --== 4. You may modify this software provided you do not remove the   ==--
REM --==    license and copyright notice.                                 ==--
REM --==                                                                  ==--
REM --== 5. You may distribute this software and derivative works to      ==--
REM --==    personal friends and work colleagues only.                    ==--
REM --==                                                                  ==--
REM --== 6. You agree that this software comes as-is and with no        ==--
REM --==    warranty whatsoever, either expressed or implied, including,  ==--
REM --==    but not limited to, warranties of merchantability or fitness  ==--
REM --==    for a particular purpose.                                     ==--
REM --==                                                                  ==--
REM --== 7. You agree that the author will not be liable for any damages  ==--
REM --==    relating from the use of this software, including direct,     ==--
REM --==    indirect, consequential or incidental. This software is used  ==--
REM --==    entirely at your own risk and should it prove defective, you  ==--
REM --==    will assume full responsibility for all costs associated with ==--
REM --==    servicing, repair or correction.                              ==--
REM --==                                                                  ==--
REM --== Your rights under this license are terminated immediately if you ==--
REM --== breach it in any way.                                            ==--
REM --==                                                                  ==--
REM ---======================= End Software License =======================---


REM ---====================== Start Copyright Notice ======================---
REM --==                                                                  ==--
REM --== Filename ..... build_retro.bat                                   ==--
REM --== Download ..... http://www.spacewire.co.uk                        ==--
REM --== Author ....... Steve Haywood (steve.haywood@ukonline.co.uk)      ==--
REM --== Copyright .... Copyright (c) 2004 Steve Haywood                  ==--
REM --== Project ...... Raptor-16                                         ==--
REM --== Version ...... 1.00                                              ==--
REM --== Conception ... 29 August 2004                                    ==--
REM --== Modified ..... N/A                                               ==--
REM --==                                                                  ==--
REM ---======================= End Copyright Notice =======================---


REM ---========================= Start Description ========================---
REM --==                                                                  ==--
REM --== Synthesis and Place & Route Script for Xilinx WebPack 6.2i.      ==--
REM --==                                                                  ==--
REM ---========================== End Description =========================---


REM ---=====================---
REM --== Design Parameters ==--
REM ---=====================---

SET part=xc3s200-4-ft256
SET design=top_level
SET prom=xcf01s

REM ---===========================================---
REM --== Delete synthesis directory if it exists ==--
REM ---===========================================---

IF EXIST synth RMDIR /S /Q synth

REM ---==================================---
REM --== Create new synthesis directory ==--
REM ---==================================---

MKDIR synth

REM ---================================---
REM --== CD into synthesis directory  ==--
REM ---================================---

CD synth

REM ---============================---
REM --== Create VHDL compile list ==--
REM ---============================---

ECHO ..\src\raptor16.vhd             > vhdl_files.prj
ECHO ..\src\memory_spartan3.vhd     >> vhdl_files.prj
ECHO ..\src\raptor16_retro_top.vhd  >> vhdl_files.prj

REM ---============================---
REM --== Create synthesise script ==--
REM ---============================---

ECHO run                     > synth.scr
ECHO -ifn vhdl_files.prj    >> synth.scr
ECHO -ifmt VHDL             >> synth.scr
ECHO -ofn %design%.ngc      >> synth.scr
ECHO -ofmt NGC              >> synth.scr
ECHO -p %part%              >> synth.scr
ECHO -opt_mode Area         >> synth.scr
ECHO -opt_level 2           >> synth.scr

REM ---=========================---
REM --== Synthesise the design ==--
REM ---=========================---

XST -ifn synth.scr

REM ---==================================---
REM --== CD out of synthesis directory  ==--
REM ---==================================---

CD ..

REM ---=========================================---
REM --== Delete fitting directory if it exists ==--
REM ---=========================================---

IF EXIST fitting rmdir /S /Q fitting

REM ---================================---
REM --== Create new fitting directory ==--
REM ---================================---

MKDIR fitting

REM ---==============================---
REM --== CD into fitting directory  ==--
REM ---==============================---

CD fitting

REM ---====================================---
REM --== Create Place & Route Constraints ==--
REM ---====================================---

ECHO NET "rst_n"            LOC = "T14";              > constraints.ucf

ECHO NET "utmi_databus16_8" LOC = "B14";             >> constraints.ucf
ECHO NET "utmi_reset"       LOC = "B4";              >> constraints.ucf
ECHO NET "utmi_xcvrselect"  LOC = "D6";              >> constraints.ucf
ECHO NET "utmi_termselect"  LOC = "B6";              >> constraints.ucf
ECHO NET "utmi_opmode1"     LOC = "C6";              >> constraints.ucf
ECHO NET "utmi_txvalid"     LOC = "B12";             >> constraints.ucf
ECHO NET "utmi_clkout"      LOC = "D9";              >> constraints.ucf

ECHO NET "lcd_rs"           LOC = "P1";              >> constraints.ucf
ECHO NET "lcd_rw"           LOC = "N3";              >> constraints.ucf
ECHO NET "lcd_en"           LOC = "N2";              >> constraints.ucf

ECHO NET "lcd_data<4>"      LOC = "N1";              >> constraints.ucf
ECHO NET "lcd_data<5>"      LOC = "L5";              >> constraints.ucf
ECHO NET "lcd_data<6>"      LOC = "L4";              >> constraints.ucf
ECHO NET "lcd_data<7>"      LOC = "M1";              >> constraints.ucf

ECHO NET "utmi_clkout" TNM_NET = "utmi_clkout";      >> constraints.ucf
ECHO TIMESPEC "ts_clk" = PERIOD "utmi_clkout" 20 ns; >> constraints.ucf

REM ---==========================---
REM --== Combine design modules ==--
REM ---==========================---

NGDBUILD -p %part% -uc constraints.ucf -dd tmp ..\synth\%design%.ngc %design%.ngd

REM ---=================================---
REM --== Map design modules into gates ==--
REM ---=================================---

MAP -detail -r -cm speed -p %part% -o map.ncd -pr b %design%.ngd %design%.pcf

REM ---===========================---
REM --== Perform place and route ==--
REM ---===========================---

PAR -w -ol 5 map.ncd %design%.ncd %design%.pcf

REM ---===================================---
REM --== Report on critical timing paths ==--
REM ---===================================---

TRCE -v 50 %design%.ncd %design%.pcf

REM ---==============================---
REM --== Create BitGen options file ==--
REM ---==============================---

ECHO -g DebugBitstream:No   >> bitgen.ut
ECHO -g Binary:no           >> bitgen.ut
ECHO -g CRC:Enable          >> bitgen.ut
ECHO -g ConfigRate:6        >> bitgen.ut
ECHO -g CclkPin:PullUp      >> bitgen.ut
ECHO -g M0Pin:PullUp        >> bitgen.ut
ECHO -g M1Pin:PullUp        >> bitgen.ut
ECHO -g M2Pin:PullUp        >> bitgen.ut
ECHO -g ProgPin:PullUp      >> bitgen.ut
ECHO -g DonePin:PullUp      >> bitgen.ut
ECHO -g TckPin:PullUp       >> bitgen.ut
ECHO -g TdiPin:PullUp       >> bitgen.ut
ECHO -g TdoPin:PullUp       >> bitgen.ut
ECHO -g TmsPin:PullUp       >> bitgen.ut
ECHO -g UnusedPin:PullDown  >> bitgen.ut
ECHO -g UserID:0xFFFFFFFF   >> bitgen.ut
ECHO -g DCMShutDown:Disable >> bitgen.ut
ECHO -g StartUpClk:CClk     >> bitgen.ut
ECHO -g DONE_cycle:4        >> bitgen.ut
ECHO -g GTS_cycle:5         >> bitgen.ut
ECHO -g GWE_cycle:6         >> bitgen.ut
ECHO -g LCK_cycle:NoWait    >> bitgen.ut
ECHO -g Match_cycle:Auto    >> bitgen.ut
ECHO -g Security:None       >> bitgen.ut
ECHO -g DonePipe:No         >> bitgen.ut
ECHO -g DriveDone:No        >> bitgen.ut

REM ---==============================---
REM --== Produce FPGA download file ==--
REM ---==============================---

BITGEN %design%.ncd  -l -w -f bitgen.ut

REM ---===================================---
REM --== Produce Flash RAM download file ==--
REM ---===================================---

PROMGEN -w -p mcs -x %prom% -u 0 %design%.bit

REM ---================================---
REM --== CD out of fitting directory  ==--
REM ---================================---

CD ..

REM ---==========---
REM --== Enjoy! ==--
REM ---==========---
