---====================== Start Software License ======================---
--==                                                                  ==--
--== This license governs the use of this software, and your use of   ==--
--== this software constitutes acceptance of this license. Agreement  ==--
--== with all points is required to use this software.                ==--
--==                                                                  ==--
--== 1. You may use this software freely for personal use.            ==--
--==                                                                  ==--
--== 2. You may use this software freely to determine feasibility for ==--
--==    commercial use.                                               ==--
--==                                                                  ==--
--== 3. You may use this software for commercial use if the author    ==--
--==    has given you written consent.                                ==--
--==                                                                  ==--
--== 4. You may modify this software provided you do not remove the   ==--
--==    license and copyright notice.                                 ==--
--==                                                                  ==--
--== 5. You may distribute this software and derivative works to      ==--
--==    personal friends and work colleagues only.                    ==--
--==                                                                  ==--
--== 6. You agree that this software comes as-is and with no        ==--
--==    warranty whatsoever, either expressed or implied, including,  ==--
--==    but not limited to, warranties of merchantability or fitness  ==--
--==    for a particular purpose.                                     ==--
--==                                                                  ==--
--== 7. You agree that the author will not be liable for any damages  ==--
--==    relating from the use of this software, including direct,     ==--
--==    indirect, consequential or incidental. This software is used  ==--
--==    entirely at your own risk and should it prove defective, you  ==--
--==    will assume full responsibility for all costs associated with ==--
--==    servicing, repair or correction.                              ==--
--==                                                                  ==--
--== Your rights under this license are terminated immediately if you ==--
--== breach it in any way.                                            ==--
--==                                                                  ==--
---======================= End Software License =======================---


---====================== Start Copyright Notice ======================---
--==                                                                  ==--
--== Filename ..... assembler.vhd                                     ==--
--== Download ..... http://www.spacewire.co.uk                        ==--
--== Author ....... Steve Haywood (steve.haywood@ukonline.co.uk)      ==--
--== Copyright .... Copyright (c) 2004 Steve Haywood                  ==--
--== Project ...... Raptor-16                                         ==--
--== Version ...... 1.00                                              ==--
--== Conception ... 10 June 2004                                      ==--
--== Modified ..... N/A                                               ==--
--==                                                                  ==--
---======================= End Copyright Notice =======================---


---========================= Start Description ========================---
--==                                                                  ==--
--== Assembler & RAM Model for the Raptor-16 CISC Microprocessor.     ==--
--==                                                                  ==--
---========================== End Description =========================---

LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE STD.TEXTIO.ALL;

ENTITY memory IS
  PORT(clk  : IN  STD_LOGIC;
       en   : IN  STD_LOGIC;
       we   : IN  STD_LOGIC;
       addr : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       din  : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       dout : OUT STD_LOGIC_VECTOR(15 DOWNTO 0)
      );
END memory;

ARCHITECTURE behave OF memory IS

---==================---
--== User Constants ==--
---==================---

CONSTANT assembly_file     : STRING(1 TO 22)          := "asm\asm_cond_tests.txt";
--CONSTANT assembly_file     : STRING(1 TO 20)          := "asm\asm_4bit_lcd.txt";
--CONSTANT assembly_file     : STRING(1 TO 22)          := "asm\asm_pacman_lcd.txt";

CONSTANT sp2_file          : STRING(1 TO 19)          := "memory_spartan2.vhd";
CONSTANT sp3_file          : STRING(1 TO 19)          := "memory_spartan3.vhd";
CONSTANT source_path       : STRING(1 TO 4)           := "src\";

CONSTANT coe_file          : STRING(1 TO 10)          := "memory.coe";
CONSTANT xco_file          : STRING(1 TO 10)          := "memory.xco";
CONSTANT coregen_path      : STRING(1 TO 8)           := "coregen\";

CONSTANT ram_size          : NATURAL RANGE 0 TO 65536 := 2048;

CONSTANT max_label_entries : NATURAL                  := 100;
CONSTANT max_label_length  : NATURAL RANGE 5 TO 25    := 15;
CONSTANT max_code_lines    : NATURAL                  := 5000;
CONSTANT max_code_length   : NATURAL                  := 128;

CONSTANT DEBUG_MSG         : BOOLEAN                  := FALSE;
CONSTANT OPT_MSG           : BOOLEAN                  := TRUE;

---=========================---
--== Instruction Constants ==--
---=========================---

CONSTANT inst_move  : NATURAL := 0;
CONSTANT inst_and   : NATURAL := 1;
CONSTANT inst_or    : NATURAL := 2;
CONSTANT inst_xor   : NATURAL := 3;
CONSTANT inst_shl   : NATURAL := 4;
CONSTANT inst_shr   : NATURAL := 5;
CONSTANT inst_add   : NATURAL := 6;
CONSTANT inst_sub   : NATURAL := 7;
CONSTANT inst_cmp   : NATURAL := 8;
CONSTANT inst_mulu  : NATURAL := 9;
CONSTANT inst_muls  : NATURAL := 10;
CONSTANT inst_divu  : NATURAL := 11;
CONSTANT inst_divs  : NATURAL := 12;
CONSTANT inst_bra   : NATURAL := 13;
CONSTANT inst_bsr   : NATURAL := 14;
CONSTANT inst_load  : NATURAL := 15;

CONSTANT inst_moveq : NATURAL := 16 + inst_move;
CONSTANT inst_andq  : NATURAL := 16 + inst_and;
CONSTANT inst_orq   : NATURAL := 16 + inst_or;
CONSTANT inst_xorq  : NATURAL := 16 + inst_xor;
CONSTANT inst_shlq  : NATURAL := 16 + inst_shl;
CONSTANT inst_shrq  : NATURAL := 16 + inst_shr;
CONSTANT inst_addq  : NATURAL := 16 + inst_add;
CONSTANT inst_subq  : NATURAL := 16 + inst_sub;
CONSTANT inst_cmpq  : NATURAL := 16 + inst_cmp;
CONSTANT inst_muluq : NATURAL := 16 + inst_mulu;
CONSTANT inst_mulsq : NATURAL := 16 + inst_muls;
CONSTANT inst_divuq : NATURAL := 16 + inst_divu;
CONSTANT inst_divsq : NATURAL := 16 + inst_divs;
CONSTANT inst_braq  : NATURAL := 16 + inst_bra;
CONSTANT inst_bsrq  : NATURAL := 16 + inst_bsr;
CONSTANT inst_loadq : NATURAL := 16 + inst_load;

CONSTANT inst_popu  : NATURAL := 246;
CONSTANT inst_popl  : NATURAL := 247;
CONSTANT inst_pushu : NATURAL := 248;
CONSTANT inst_pushl : NATURAL := 249;
CONSTANT inst_rti   : NATURAL := 250;
CONSTANT inst_not   : NATURAL := 251;
CONSTANT inst_rts   : NATURAL := 252;
CONSTANT inst_nop   : NATURAL := 253;
CONSTANT inst_bytes : NATURAL := 254;
CONSTANT inst_words : NATURAL := 255;

---=============================---
--== Addressing Mode Constants ==--
---=============================---

CONSTANT mode_dir_lower : STD_LOGIC_VECTOR(2 DOWNTO 0) := "000";
CONSTANT mode_dir_upper : STD_LOGIC_VECTOR(2 DOWNTO 0) := "001";
CONSTANT mode_indir     : STD_LOGIC_VECTOR(2 DOWNTO 0) := "010";
CONSTANT mode_indir_pd  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "011";
CONSTANT mode_indir_pi  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "100";
CONSTANT mode_indir_dsp : STD_LOGIC_VECTOR(2 DOWNTO 0) := "101";
CONSTANT mode_imm_quick : STD_LOGIC_VECTOR(2 DOWNTO 0) := "110";
CONSTANT mode_extend    : STD_LOGIC_VECTOR(2 DOWNTO 0) := "111";

CONSTANT mode_ext_imm   : STD_LOGIC_VECTOR(2 DOWNTO 0) := "110";
CONSTANT mode_ext_abs   : STD_LOGIC_VECTOR(2 DOWNTO 0) := "111";

---=======================---
--== Mode Mask Constants ==--
---=======================---

CONSTANT mask_dir       : NATURAL := 0; -- Rn
CONSTANT mask_indir     : NATURAL := 1; -- (Rn)
CONSTANT mask_indir_pd  : NATURAL := 2; -- -(Rn)
CONSTANT mask_indir_pi  : NATURAL := 3; -- (Rn)+
CONSTANT mask_indir_dsp : NATURAL := 4; -- d(Rn) or Label(Rn)
CONSTANT mask_imm_int   : NATURAL := 5; -- #n
CONSTANT mask_abs_int   : NATURAL := 6; -- n
CONSTANT mask_abs_lab   : NATURAL := 7; -- Label
CONSTANT mask_imm_lab   : NATURAL := 8; -- #Label
CONSTANT mask_port      : NATURAL := 9; -- Pn
CONSTANT mask_control   : NATURAL := 10; -- CR

---==============================---
--== Branch Condition Constants ==--
---==============================---

CONSTANT branch_alw : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0000";
CONSTANT branch_nev : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0001";
CONSTANT branch_zrs : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0010";
CONSTANT branch_zrc : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0011";
CONSTANT branch_ngs : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0100";
CONSTANT branch_ngc : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0101";
CONSTANT branch_uos : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0110";
CONSTANT branch_uoc : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0111";
CONSTANT branch_sos : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1000";
CONSTANT branch_soc : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1001";
CONSTANT branch_ugt : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1010";
CONSTANT branch_ule : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1011";
CONSTANT branch_sgt : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1100";
CONSTANT branch_sge : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1101";
CONSTANT branch_slt : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1110";
CONSTANT branch_sle : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1111";

CONSTANT branch_equ : STD_LOGIC_VECTOR(3 DOWNTO 0) := branch_zrs;
CONSTANT branch_neq : STD_LOGIC_VECTOR(3 DOWNTO 0) := branch_zrc;
CONSTANT branch_ult : STD_LOGIC_VECTOR(3 DOWNTO 0) := branch_uos;
CONSTANT branch_uge : STD_LOGIC_VECTOR(3 DOWNTO 0) := branch_uoc;

---========================---
--== Label Type Constants ==--
---========================---

CONSTANT lab_none    : INTEGER := -1;  -- No Label
CONSTANT lab_abs     : INTEGER := 00;  -- Absolute Addressing (0 to 65535)
CONSTANT lab_braq    : INTEGER := 01;  -- Branch (Quick) displacement (-64 to 127)
CONSTANT lab_bra     : INTEGER := 02;  -- Branch displacement (-32768 to 32767) with opt!
CONSTANT lab_immq1   : INTEGER := 03;  -- Quick data (-8 to 7)
CONSTANT lab_immq2   : INTEGER := 04;  -- Quick data (0 to 15)
CONSTANT lab_imm1    : INTEGER := 05;  -- Immediate Data with Optimisation on (-8 to 7) move/and/or/xor/cmp/muls/divs
CONSTANT lab_imm2    : INTEGER := 06;  -- Immediate Data with Optimisation on (0 to 15) shl/shr/add/sub/mulu/divu
CONSTANT lab_braqimm : INTEGER := 07;  -- Branch (Quick) Immediate (0 to 7)
CONSTANT lab_imm3    : INTEGER := 08;  -- Immediate Data with Optimisation on (0 to 7) bra/bsr
CONSTANT lab_dsp     : INTEGER := 09;  -- PC Relative (-32768 to 32767)
CONSTANT lab_load    : INTEGER := 10;  -- Immediate Data (-127 to 128) load


---=====================---
--== Type Declarations ==--
---=====================---

TYPE CODE_LINES IS ARRAY(1 TO max_code_lines) OF STRING(1 TO max_code_length);
TYPE MEM IS ARRAY(0 TO ram_size-1) OF STD_LOGIC_VECTOR(15 DOWNTO 0);

TYPE LABEL_ENTRY IS RECORD
  Name  : STRING(1 TO max_label_length);
  Value : INTEGER;
END RECORD;

TYPE LABEL_ARRAY IS ARRAY(1 TO max_label_entries) OF LABEL_ENTRY;

TYPE LABELS IS RECORD
  List    : LABEL_ARRAY;
  Count   : NATURAL;
END RECORD;

TYPE INSTRUCTION_ENTRY IS RECORD
  Inst    : STD_LOGIC_VECTOR(15 DOWNTO 0);  -- Binary
  Type_1  : INTEGER;                        -- Label Type
  Label_1 : STRING(1 TO max_label_length);  -- Label Name
  Line    : NATURAL;                        -- Source Line Number
END RECORD;

TYPE INSTRUCTION_ARRAY IS ARRAY(0 TO ram_size-1) OF INSTRUCTION_ENTRY;

TYPE INSTRUCTIONS IS RECORD
  List : INSTRUCTION_ARRAY;
  Ptr  : NATURAL;
END RECORD;

---=======================---
--== Signal Declarations ==--
---=======================---

SIGNAL memory     : MEM := (OTHERS => "1111111111111111");
SIGNAL memory_tmp : MEM := (OTHERS => "1111111111111111");

SIGNAL dout_i : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');

---==========================---
--== Functions & Procedures ==--
---==========================---

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE swallow_spaces(str : IN    STRING;     -- Line of source code
                         len : IN    NATURAL;    -- Length of line
                         pos : INOUT NATURAL) IS -- Character position
BEGIN
  IF (pos <= len) AND (str(pos) = ' ') THEN
    WHILE ((pos <= len) AND (str(pos) = ' ')) LOOP
      pos := pos + 1;
    END LOOP;
  END IF;
END swallow_spaces;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE err_msg(str   : IN STRING;    -- Line of source code
                  ln    : IN NATURAL;   -- Length of line
                  pos   : IN NATURAL;   -- Character position
                  err   : IN STRING) IS -- Error string

VARIABLE line_out : LINE;

BEGIN
  WRITE(line_out, STRING'("**"));
  WRITELINE(OUTPUT, line_out);
  WRITE(line_out, STRING'("** Problem with assembler file "));
  WRITE(line_out, assembly_file);
  WRITE(line_out, STRING'(" on line "));
  WRITE(line_out, ln);
  IF (pos > 0) THEN
    WRITE(line_out, STRING'(" at character "));
    WRITE(line_out, pos);
  END IF;
  WRITE(line_out, STRING'("."));
  WRITELINE(OUTPUT, line_out);
  WRITE(line_out, STRING'("**"));
  WRITELINE(OUTPUT, line_out);
  WRITE(line_out, STRING'("** > "));
  WRITE(line_out, str);
  WRITELINE(OUTPUT, line_out);
  WRITE(line_out, STRING'("**"));
  WRITELINE(OUTPUT, line_out);
  WRITE(line_out, STRING'("** "));
  WRITE(line_out, err);
  WRITELINE(OUTPUT, line_out);
  WRITE(line_out, STRING'("**"));
  WRITELINE(OUTPUT, line_out);
  ASSERT FALSE SEVERITY FAILURE;
END err_msg;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE optim_msg(str   : IN STRING;    -- Line of source code
                    ln    : IN NATURAL;   -- Length of line
                    pos   : IN NATURAL;   -- Character position
                    opt   : IN STRING) IS -- Warning string

VARIABLE line_out : LINE;

BEGIN
  IF (OPT_MSG) THEN
    WRITE(line_out, STRING'("**"));
    WRITELINE(OUTPUT, line_out);
    WRITE(line_out, STRING'("** Optimisation possible in assembler file "));
    WRITE(line_out, assembly_file);
    WRITE(line_out, STRING'(" on line "));
    WRITE(line_out, ln);
    IF (pos > 0) THEN
      WRITE(line_out, STRING'(" at character "));
      WRITE(line_out, pos);
    END IF;
    WRITE(line_out, STRING'("."));
    WRITELINE(OUTPUT, line_out);
    WRITE(line_out, STRING'("**"));
    WRITELINE(OUTPUT, line_out);
    WRITE(line_out, STRING'("** > "));
    WRITE(line_out, str);
    WRITELINE(OUTPUT, line_out);
    WRITE(line_out, STRING'("**"));
    WRITELINE(OUTPUT, line_out);
    WRITE(line_out, STRING'("** "));
    WRITE(line_out, opt);
    WRITELINE(OUTPUT, line_out);
    WRITE(line_out, STRING'("**"));
    WRITELINE(OUTPUT, line_out);
  END IF;
END optim_msg;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE add_to_ram(str      : IN    STRING;                        -- Line of source code
                     ln       : IN    NATURAL;                       -- Line number
                     pos      : INOUT NATURAL;                       -- Character position
                     inst     : IN    STD_LOGIC_VECTOR(15 DOWNTO 0); -- Instruction/Data
                     lab      : IN    STRING;                        -- Label Name
                     typ      : IN    INTEGER;                       -- Label Type
                     inst_rec : INOUT INSTRUCTIONS) IS               -- Instruction Record

  VARIABLE line_out : LINE;
  VARIABLE index3 : NATURAL;

BEGIN
  IF (inst_rec.ptr < ram_size) THEN
    IF DEBUG_MSG THEN
      WRITE(line_out, inst_rec.ptr, right, 4);
      WRITE(line_out, STRING'(" - "));
      WRITE(line_out, TO_BITVECTOR(inst));
      WRITE(line_out, STRING'(" - "));
      WRITE(line_out, ln, right, 4);
      IF (inst_rec.ptr = 0) THEN
        WRITE(line_out, STRING'(" - "));
        WRITE(line_out, str);
      ELSIF (inst_rec.list(inst_rec.ptr-1).line = ln) THEN
        WRITE(line_out, STRING'(" +"));
      ELSE
        WRITE(line_out, STRING'(" - "));
        WRITE(line_out, str);
      END IF;
      WRITELINE(OUTPUT, line_out);
    END IF;
    inst_rec.list(inst_rec.ptr).inst    := inst;
    inst_rec.list(inst_rec.ptr).type_1  := typ;
    inst_rec.list(inst_rec.ptr).label_1(1 TO lab'LENGTH) := lab;
    inst_rec.list(inst_rec.ptr).line    := ln;
    inst_rec.ptr := inst_rec.ptr + 1;
  ELSE
    err_msg(str, ln, pos, "RAM full, reduce/optimise code or increase RAM size.");
  END IF;
END add_to_ram;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE check_line_end(str : IN    STRING;     -- Line of source code
                         len : IN    NATURAL;    -- Length of line
                         ln  : IN    NATURAL;    -- Line number
                         pos : INOUT NATURAL) IS -- Character position
BEGIN
  IF (pos <= len) THEN
    IF (str(pos) = ' ') THEN
      WHILE ((pos <= len) AND (str(pos) = ' ')) LOOP
        pos := pos + 1;
      END LOOP;
      IF (pos <= len) THEN
        IF NOT(str(pos) = ';') THEN
          err_msg(str, ln, pos, "Junk found at end of line.");
        END IF;
      END IF;
    ELSE
      err_msg(str, ln, pos, "Junk found at end of line.");
    END IF;
  END IF;
END check_line_end;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

FUNCTION vec2char(vec : STD_LOGIC_VECTOR(3 DOWNTO 0)) RETURN CHARACTER IS
BEGIN
  CASE vec IS
    WHEN "0000" => RETURN '0';
    WHEN "0001" => RETURN '1';
    WHEN "0010" => RETURN '2';
    WHEN "0011" => RETURN '3';
    WHEN "0100" => RETURN '4';
    WHEN "0101" => RETURN '5';
    WHEN "0110" => RETURN '6';
    WHEN "0111" => RETURN '7';
    WHEN "1000" => RETURN '8';
    WHEN "1001" => RETURN '9';
    WHEN "1010" => RETURN 'A';
    WHEN "1011" => RETURN 'B';
    WHEN "1100" => RETURN 'C';
    WHEN "1101" => RETURN 'D';
    WHEN "1110" => RETURN 'E';
    WHEN "1111" => RETURN 'F';
    WHEN OTHERS => RETURN '0';
  END CASE;
END vec2char;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

FUNCTION upcase(char : CHARACTER) RETURN CHARACTER IS
BEGIN
  IF (char >= 'a') AND (char <= 'z') THEN
    RETURN CHARACTER'VAL(CHARACTER'POS(char) - 32);
  ELSE
    RETURN char;
  END IF;
END upcase;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE get_integer(str     : IN    STRING;     -- Line of source code
                      len     : IN    NATURAL;    -- Length of line
                      pos     : INOUT NATURAL;    -- Character position
                      lowest  : IN    INTEGER;    -- Lowest value allowed
                      highest : IN    INTEGER;    -- Highest value allowed
                      int     : OUT   INTEGER;    -- Return : Integer
                      err     : OUT   NATURAL) IS -- Return : Error

  vARIABLE got_digit : BOOLEAN := FALSE;
  VARIABLE negative  : BOOLEAN := FALSE;
  VARIABLE int_i     : INTEGER := 0;
  VARIABLE err_i     : NATURAL := 0;

BEGIN
  IF (pos <= len) THEN
    IF (str(pos) = '%') THEN
      pos := pos + 1;
      WHILE (pos <= len) AND ((str(pos) = '0') OR (str(pos) = '1')) AND (err_i = 0) LOOP
        got_digit := TRUE;
        int_i := int_i*2 + (CHARACTER'POS(upcase(str(pos))) - CHARACTER'POS('0'));
        IF (int_i > 65535) THEN -- Hard overflow
          err_i := 2;
        END IF;
        pos := pos + 1;
      END LOOP;
    ELSIF (str(pos) = '$') THEN
      pos := pos + 1;
      WHILE (pos <= len) AND (((str(pos) >= '0') AND (str(pos) <= '9')) OR
            ((upcase(str(pos)) >= 'A') AND (upcase(str(pos)) <= 'F'))) AND (err_i = 0) LOOP
        got_digit := TRUE;
        IF (str(pos) >= '0') AND (str(pos) <= '9') THEN
          int_i := int_i*16 + (CHARACTER'POS(str(pos)) - CHARACTER'POS('0'));
        ELSE
          int_i := int_i*16 + 10 + (CHARACTER'POS(upcase(str(pos))) - CHARACTER'POS('A'));
        END IF;
        IF (int_i > 65535) THEN -- Hard overflow
          err_i := 2;
        END IF;
        pos := pos + 1;
      END LOOP;
    ELSIF ((str(pos) >= '0') AND (str(pos) <= '9')) OR (str(pos) = '-') THEN
      IF (pos <= len) AND (str(pos) = '-') THEN
        negative := TRUE;
        pos := pos + 1;
      END IF;
      WHILE ((pos <= len) AND (str(pos) >= '0') AND (str(pos) <= '9') AND (err_i = 0)) LOOP
        got_digit := TRUE;
        IF negative THEN
          int_i := int_i*10 - (CHARACTER'POS(str(pos)) - CHARACTER'POS('0'));
        ELSE
          int_i := int_i*10 + (CHARACTER'POS(str(pos)) - CHARACTER'POS('0'));
        END IF;
        IF (int_i < -32768) OR (int_i > 65535) THEN -- Hard overflow
          err_i := 2;
        END IF;
        pos := pos + 1;
      END LOOP;
    ELSE
      err_i := 1; -- No Number
    END IF;
    IF NOT(got_digit) THEN
      err_i := 1; -- No Number
    ELSE
      IF (int_i < lowest) OR (int_i > highest) THEN -- Soft overflow
        err_i := 2;
      END IF;
    END IF;
  ELSE
    err_i := 1; -- No Number
  END IF;
  int := int_i;
  err := err_i;
END get_integer;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE get_label(str  : IN    STRING;                           -- Line of source code
                    len  : IN    NATURAL;                          -- Length of line
                    ln   : IN    NATURAL;                          -- Line number
                    pos  : INOUT NATURAL;                          -- Character position
                    name : OUT   STRING(1 TO max_label_length)) IS -- Return : Label

  VARIABLE name_i     : STRING(1 TO max_label_length) := (OTHERS => NUL);
  VARIABLE char_count : NATURAL := 0;

BEGIN
  WHILE (pos <= len) AND (upcase(str(pos)) >= 'A') AND (upcase(str(pos)) <= 'Z') LOOP
    IF (char_count < max_label_length) THEN
      char_count := char_count + 1;
      name_i(char_count) := upcase(str(pos));
      pos := pos + 1;
    ELSE
      err_msg(str, ln, pos, "Label too long.");
    END IF;
  END LOOP;
  IF (char_count = 0) THEN
    err_msg(str, ln, pos, "Label missing or invalid (use characters a-z and A-Z only).");
  END IF;
  name := name_i;
END get_label;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE get_operand(str     : IN    STRING;                           -- Line of source code
                      len     : IN    NATURAL;                          -- Length of line
                      ln      : IN    NATURAL;                          -- Line number
                      pos     : INOUT NATURAL;                          -- Character position
                      mask    : IN    STD_LOGIC_VECTOR(10 DOWNTO 0);    -- Mode Mask
                      mode    : OUT   STD_LOGIC_VECTOR(2 DOWNTO 0);     -- Return : Adressing Mode
                      reg     : OUT   STD_LOGIC_VECTOR(2 DOWNTO 0);     -- Return : Register
                      data    : OUT   STD_LOGIC_VECTOR(15 DOWNTO 0);    -- Return : Data (Disp/Abs/Imm)
                      lab     : OUT   STRING(1 TO max_label_length)) IS -- Return : Label (Abs/Imm)

  VARIABLE lab_i   : STRING(1 TO max_label_length) := (OTHERS => NUL);
  VARIABLE int     : INTEGER;
  VARIABLE err     : NATURAL;
  VARIABLE old_pos : NATURAL;

BEGIN
  IF (pos <= len) THEN

    IF (upcase(str(pos)) = 'C') AND (upcase(str(pos+1)) = 'R') AND
          ((str(pos+2) = ' ') OR (str(pos+2) = ',') OR (str(pos+2) = NUL)) THEN
      IF (mask(mask_control) = '1') THEN
        mode := mode_extend;
        reg := "101";
        pos := pos + 2;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;

    ELSIF (upcase(str(pos)) = 'P') AND (str(pos+1) >= '0') AND (str(pos+1) <= '4') AND
       ((str(pos+2) = ' ') OR (str(pos+2) = ',') OR (str(pos+2) = NUL)) THEN
      IF (mask(mask_port) = '1') THEN
        mode := mode_extend;
        reg := STD_LOGIC_VECTOR(CONV_UNSIGNED((CHARACTER'POS(upcase(str(pos+1))) - CHARACTER'POS('0')), 3));
        pos := pos + 2;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (upcase(str(pos)) = 'D') AND (str(pos+1) >= '0') AND (str(pos+1) <= '7') AND
          ((str(pos+2) = ' ') OR (str(pos+2) = ',') OR (str(pos+2) = NUL)) THEN
      IF (mask(mask_dir) = '1') THEN
        mode := mode_dir_lower;
        reg := STD_LOGIC_VECTOR(CONV_UNSIGNED((CHARACTER'POS(upcase(str(pos+1))) - CHARACTER'POS('0')), 3));
        pos := pos + 2;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (upcase(str(pos)) = 'A') AND (str(pos+1) >= '0') AND (str(pos+1) <= '5') AND
          ((str(pos+2) = ' ') OR (str(pos+2) = ',') OR (str(pos+2) = NUL)) THEN
      IF (mask(mask_dir) = '1') THEN
        mode := mode_dir_upper;
        reg := STD_LOGIC_VECTOR(CONV_UNSIGNED((CHARACTER'POS(upcase(str(pos+1))) - CHARACTER'POS('0')), 3));
        pos := pos + 2;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (upcase(str(pos)) = 'S') AND (upcase(str(pos+1)) = 'P') AND
          ((str(pos+2) = ' ') OR (str(pos+2) = ',') OR (str(pos+2) = NUL)) THEN
      IF (mask(mask_dir) = '1') THEN
        mode := mode_dir_upper;
        reg := "110";
        pos := pos + 2;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (upcase(str(pos)) = 'P') AND (upcase(str(pos+1)) = 'C') AND
          ((str(pos+2) = ' ') OR (str(pos+2) = ',') OR (str(pos+2) = NUL)) THEN
      IF (mask(mask_dir) = '1') THEN
        mode := mode_dir_upper;
        reg := "111";
        pos := pos + 2;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (str(pos) = '(') AND (upcase(str(pos+1)) = 'A') AND (str(pos+2) >= '0') AND (str(pos+2) <= '5') AND (str(pos+3) = ')') AND
          ((str(pos+4) = ' ') OR (str(pos+4) = ',') OR (str(pos+4) = NUL)) THEN
      IF (mask(mask_indir) = '1') THEN
        mode := mode_indir;
        reg := STD_LOGIC_VECTOR(CONV_UNSIGNED((CHARACTER'POS(upcase(str(pos+2))) - CHARACTER'POS('0')), 3));
        pos := pos + 4;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (str(pos) = '(') AND (upcase(str(pos+1)) = 'S') AND (upcase(str(pos+2)) = 'P') AND (str(pos+3) = ')') AND
          ((str(pos+4) = ' ') OR (str(pos+4) = ',') OR (str(pos+4) = NUL)) THEN
      IF (mask(mask_indir) = '1') THEN
        mode := mode_indir;
        reg := "110";
        pos := pos + 4;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (str(pos) = '(') AND (upcase(str(pos+1)) = 'P') AND (upcase(str(pos+2)) = 'C') AND (str(pos+3) = ')') AND
          ((str(pos+4) = ' ') OR (str(pos+4) = ',') OR (str(pos+4) = NUL)) THEN
      IF (mask(mask_indir) = '1') THEN
        mode := mode_indir;
        reg := "111";
        pos := pos + 4;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;

    ELSIF (str(pos) = '(') AND (upcase(str(pos+1)) = 'A') AND (str(pos+2) >= '0') AND (str(pos+2) <= '5') AND (str(pos+3) = ')') AND
          (str(pos+4) = '+') AND
          ((str(pos+5) = ' ') OR (str(pos+5) = ',') OR (str(pos+5) = NUL)) THEN
      IF (mask(mask_indir_pi) = '1') THEN
        mode := mode_indir_pi;
        reg := STD_LOGIC_VECTOR(CONV_UNSIGNED((CHARACTER'POS(upcase(str(pos+2))) - CHARACTER'POS('0')), 3));
        pos := pos + 5;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (str(pos) = '(') AND (upcase(str(pos+1)) = 'S') AND (upcase(str(pos+2)) = 'P') AND (str(pos+3) = ')') AND
          (str(pos+4) = '+') AND
          ((str(pos+5) = ' ') OR (str(pos+5) = ',') OR (str(pos+5) = NUL)) THEN
      IF (mask(mask_indir_pi) = '1') THEN
        mode := mode_indir_pi;
        reg := "110";
        pos := pos + 5;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (str(pos) = '(') AND (upcase(str(pos+1)) = 'P') AND (upcase(str(pos+2)) = 'C') AND (str(pos+3) = ')') AND
          (str(pos+4) = '+') AND
          ((str(pos+5) = ' ') OR (str(pos+5) = ',') OR (str(pos+5) = NUL)) THEN
      IF (mask(mask_indir_pi) = '1') THEN
        mode := mode_indir_pi;
        reg := "111";
        pos := pos + 5;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;

    ELSIF (str(pos) = '-') AND (str(pos+1) = '(') AND (upcase(str(pos+2)) = 'A') AND (str(pos+3) >= '0') AND (str(pos+3) <= '5') AND
          (str(pos+4) = ')') AND
          ((str(pos+5) = ' ') OR (str(pos+5) = ',') OR (str(pos+5) = NUL)) THEN
      IF (mask(mask_indir_pd) = '1') THEN
        mode := mode_indir_pd;
        reg := STD_LOGIC_VECTOR(CONV_UNSIGNED((CHARACTER'POS(upcase(str(pos+3))) - CHARACTER'POS('0')), 3));
        pos := pos + 5;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (str(pos) = '-') AND (str(pos+1) = '(') AND (upcase(str(pos+2)) = 'S') AND (upcase(str(pos+3)) = 'P') AND
          (str(pos+4) = ')') AND
          ((str(pos+5) = ' ') OR (str(pos+5) = ',') OR (str(pos+5) = NUL)) THEN
      IF (mask(mask_indir_pd) = '1') THEN
        mode := mode_indir_pd;
        reg := "110";
        pos := pos + 5;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (str(pos) = '-') AND (str(pos+1) = '(') AND (upcase(str(pos+2)) = 'P') AND (upcase(str(pos+3)) = 'C') AND
          (str(pos+4) = ')') AND
          ((str(pos+5) = ' ') OR (str(pos+5) = ',') OR (str(pos+5) = NUL)) THEN
      IF (mask(mask_indir_pd) = '1') THEN
        mode := mode_indir_pd;
        reg := "111";
        pos := pos + 5;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF ((str(pos) >= '0') AND (str(pos) <= '9')) OR
          (str(pos) = '-') OR (str(pos) = '$') OR (str(pos) = '%') THEN -- d(An) or Abs
      old_pos := pos;
      IF (mask(mask_indir_dsp) = '1') OR (mask(mask_abs_int) = '1') THEN
        get_integer(str, len, pos, -32768, 65535, int, err);
        IF (err = 0) THEN
          IF (str(pos) = '(') AND (upcase(str(pos+1)) = 'A') AND (str(pos+2) >= '0') AND (str(pos+2) <= '5') AND (str(pos+3) = ')') AND
             ((str(pos+4) = ' ') OR (str(pos+4) = ',') OR (str(pos+4) = NUL)) THEN -- d(An)
            IF (mask(mask_indir_dsp) = '1') THEN
              IF ((str(old_pos) = '$') OR (str(old_pos) = '%')) AND (int > 32767) THEN -- Make Negative
                int := int - 65536;
              END IF;
              IF (int <= 32767) THEN
                data := STD_LOGIC_VECTOR(CONV_SIGNED(int, 16));
                mode := mode_indir_dsp;
                reg := STD_LOGIC_VECTOR(CONV_UNSIGNED((CHARACTER'POS(upcase(str(pos+2))) - CHARACTER'POS('0')), 3));
                pos := pos + 4;
              ELSE
                err_msg(str, ln, pos, "Operand displacement out of range (use -32768..32767).");
              END IF;
            ELSE
              err_msg(str, ln, pos, "Operand corrupt.");
            END IF;
          ELSIF (str(pos) = '(') AND (upcase(str(pos+1)) = 'S') AND (upcase(str(pos+2)) = 'P') AND (str(pos+3) = ')') AND
                ((str(pos+4) = ' ') OR (str(pos+4) = ',') OR (str(pos+4) = NUL)) THEN -- d(SP)
            IF (mask(mask_indir_dsp) = '1') THEN
              IF ((str(old_pos) = '$') OR (str(old_pos) = '%')) AND (int > 32767) THEN -- Make Negative
                int := int - 65536;
              END IF;
              IF (int <= 32767) THEN
                data := STD_LOGIC_VECTOR(CONV_SIGNED(int, 16));
                mode := mode_indir_dsp;
                reg := "110";
                pos := pos + 4;
              ELSE
                err_msg(str, ln, pos, "Operand displacement out of range (use -32768..32767).");
              END IF;
            ELSE
              err_msg(str, ln, pos, "Operand corrupt.");
            END IF;
          ELSIF (str(pos) = '(') AND (upcase(str(pos+1)) = 'P') AND (upcase(str(pos+2)) = 'C') AND (str(pos+3) = ')') AND
                ((str(pos+4) = ' ') OR (str(pos+4) = ',') OR (str(pos+4) = NUL)) THEN -- d(PC)
            IF (mask(mask_indir_dsp) = '1') THEN
              IF ((str(old_pos) = '$') OR (str(old_pos) = '%')) AND (int > 32767) THEN -- Make Negative
                int := int - 65536;
              END IF;
              IF (int <= 32767) THEN
                data := STD_LOGIC_VECTOR(CONV_SIGNED(int, 16));
                mode := mode_indir_dsp;
                reg := "111";
                pos := pos + 4;
              ELSE
                err_msg(str, ln, pos, "Operand displacement out of range (use -32768..32767).");
              END IF;
            ELSE
              err_msg(str, ln, pos, "Operand corrupt.");
            END IF;
          ELSIF ((str(pos) = ' ') OR (str(pos) = ',') OR (str(pos) = NUL)) THEN -- Abs
            IF (mask(mask_abs_int) = '1') THEN
              IF (int >= 0) THEN
                mode := mode_extend;
                reg := mode_ext_abs;
                data := STD_LOGIC_VECTOR(CONV_UNSIGNED(int, 16));
              ELSE
                err_msg(str, ln, pos, "Operand absolute address out of range (use 0..65535).");
              END IF;
            ELSE
              err_msg(str, ln, pos, "Operand corrupt.");
            END IF;
          ELSE
            err_msg(str, ln, pos, "Operand corrupt.");
          END IF;
        ELSE
          IF (err = 1) THEN
            err_msg(str, ln, pos, "Operand displacement or absolute address missing.");
          ELSE
            err_msg(str, ln, pos, "Operand displacement or absolute address out of range.");
          END IF;
        END IF;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;
    ELSIF (upcase(str(pos)) >= 'A') AND (upcase(str(pos)) <= 'Z') THEN -- Label

      get_label(str, len, ln, pos, lab_i);
      IF (str(pos) = '(') AND (upcase(str(pos+1)) = 'P') AND (upcase(str(pos+2)) = 'C') AND (str(pos+3) = ')') AND
         ((str(pos+4) = ' ') OR (str(pos+4) = ',') OR (str(pos+4) = NUL)) THEN -- Label(PC)
        IF (mask(mask_indir_dsp) = '1') THEN
          mode := mode_indir_dsp;
          reg := "111";
          pos := pos + 4;
        ELSE
          err_msg(str, ln, pos, "Operand corrupt.");
        END IF;
      ELSIF ((str(pos) = ' ') OR (str(pos) = ',') OR (str(pos) = NUL)) THEN -- Abs Label
        IF (mask(mask_abs_lab) = '1') THEN
          mode := mode_extend;
          reg := mode_ext_abs;
        ELSE
          err_msg(str, ln, pos, "Operand corrupt.");
        END IF;
      ELSE
        err_msg(str, ln, pos, "Operand corrupt.");
      END IF;

    ELSIF (str(pos) = '#') THEN -- Imm
      pos := pos + 1;
      IF (pos <= len) THEN
        IF (str(pos) = '-') OR ((str(pos) >= '0') AND (str(pos) <= '9')) OR
           (str(pos) = '$') OR (str(pos) = '%') THEN -- Integer
          IF (mask(mask_imm_int) = '1') THEN
            get_integer(str, len, pos, -32768, 65535, int, err);
            IF (err = 0) THEN
              mode := mode_extend;
              reg := mode_ext_imm;
              IF (int < 0) THEN
                data := STD_LOGIC_VECTOR(CONV_SIGNED(int, 16));
              ELSE
                data := STD_LOGIC_VECTOR(CONV_UNSIGNED(int, 16));
              END IF;
            ELSE
              IF (err = 1) THEN
                err_msg(str, ln, pos, "Operand immediate value missing.");
              ELSE
                err_msg(str, ln, pos, "Operand immediate value out of range (use -32768..32767 or 0..65535).");
              END IF;
            END IF;
          ELSE
            err_msg(str, ln, pos, "Operand corrupt.");
          END IF;
        ELSIF (upcase(str(pos)) >= 'A') AND (upcase(str(pos)) <= 'Z') THEN -- Label
          IF (mask(mask_imm_lab) = '1') THEN
            mode := mode_extend;
            reg := mode_ext_imm;
            get_label(str, len, ln, pos, lab_i);
          ELSE
            err_msg(str, ln, pos, "Operand corrupt.");
          END IF;
        ELSE
          err_msg(str, ln, pos, "Operand immediate value invalid or corrupt.");
        END IF;
      ELSE
        err_msg(str, ln, pos, "Operand immediate value missing.");
      END IF;
    ELSE
      err_msg(str, ln, pos, "Operand corrupt.");
    END IF;
  ELSE
    err_msg(str, ln, pos, "Operand missing.");
  END IF;
  lab := lab_i;
END get_operand;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE get_instruction(str     : IN    STRING;     -- Line of source code
                          len     : IN    NATURAL;    -- Length of line
                          ln      : IN    NATURAL;    -- Line number
                          pos     : INOUT NATURAL;    -- Character position
                          inst    : OUT   NATURAL) IS -- Return : Instruction

  VARIABLE inst_str : STRING(1 TO 255) := (OTHERS => NUL);
  VARIABLE inst_len : NATURAL := 0;
  VARIABLE inst_i   : NATURAL := 0;

BEGIN
  IF (pos <= len) THEN
    WHILE (pos <= len) AND (upcase(str(pos)) >= 'A') AND (upcase(str(pos)) <= 'Z') LOOP
      inst_len := inst_len + 1;
      inst_str(inst_len) := upcase(str(pos));
      pos := pos + 1;
    END LOOP;

    IF (pos <= len) AND NOT((str(pos) = ' ')) THEN
      err_msg(str, ln, pos, "Corrupt or invalid instruction.");
    END IF;

    swallow_spaces(str, len, pos);

    IF (inst_len = 4) AND (inst_str(1 TO 4) = "MOVE") THEN
      inst_i := inst_move;
    ELSIF (inst_len = 5) AND (inst_str(1 TO 5) = "MOVEQ") THEN
      inst_i := inst_moveq;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "AND") THEN
      inst_i := inst_and;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "ANDQ") THEN
      inst_i := inst_andq;
    ELSIF (inst_len = 2) AND (inst_str(1 TO 2) = "OR") THEN
      inst_i := inst_or;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "ORQ") THEN
      inst_i := inst_orq;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "XOR") THEN
      inst_i := inst_xor;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "XORQ") THEN
      inst_i := inst_xorq;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "SHL") THEN
      inst_i := inst_shl;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "SHLQ") THEN
      inst_i := inst_shlq;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "SHR") THEN
      inst_i := inst_shr;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "SHRQ") THEN
      inst_i := inst_shrq;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "ADD") THEN
      inst_i := inst_add;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "ADDQ") THEN
      inst_i := inst_addq;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "SUB") THEN
      inst_i := inst_sub;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "SUBQ") THEN
      inst_i := inst_subq;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "CMP") THEN
      inst_i := inst_cmp;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "CMPQ") THEN
      inst_i := inst_cmpq;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "BRA") THEN
      inst_i := inst_bra;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "BRAQ") THEN
      inst_i := inst_braq;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "BSR") THEN
      inst_i := inst_bsr;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "BSRQ") THEN
      inst_i := inst_bsrq;
    ELSIF (inst_len = 5) AND (inst_str(1 TO 5) = "LOADQ") THEN
      inst_i := inst_load;

    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "POPU") THEN
      inst_i := inst_popu;
    ELSIF (inst_len = 4) AND (inst_str(1 TO 4) = "POPL") THEN
      inst_i := inst_popl;
    ELSIF (inst_len = 5) AND (inst_str(1 TO 5) = "PUSHU") THEN
      inst_i := inst_pushu;
    ELSIF (inst_len = 5) AND (inst_str(1 TO 5) = "PUSHL") THEN
      inst_i := inst_pushl;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "RTI") THEN
      inst_i := inst_rti;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "NOT") THEN
      inst_i := inst_not;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "RTS") THEN
      inst_i := inst_rts;
    ELSIF (inst_len = 3) AND (inst_str(1 TO 3) = "NOP") THEN
      inst_i := inst_nop;
    ELSIF (inst_len = 5) AND (inst_str(1 TO 5) = "BYTES") THEN
      inst_i := inst_bytes;
    ELSIF (inst_len = 5) AND (inst_str(1 TO 5) = "WORDS") THEN
      inst_i := inst_words;
    ELSE
      err_msg(str, ln, pos, "Unknown instruction.");
    END IF;
  ELSE
    err_msg(str, ln, pos, "Data or instruction missing.");
  END IF;
  inst := inst_i;
END get_instruction;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE compile_words(str      : IN    STRING;          -- Line of source code
                        len      : IN    NATURAL;         -- Length of line
                        ln       : IN    NATURAL;         -- Line number
                        pos      : INOUT NATURAL;         -- Character position
                        inst_rec : INOUT INSTRUCTIONS) IS -- Instruction Record

  VARIABLE got_comma     : BOOLEAN := TRUE;
  VARIABLE found_comment : BOOLEAN := FALSE;
  VARIABLE int           : INTEGER;
  VARIABLE err           : NATURAL;
  VARIABLE data          : STD_LOGIC_VECTOR(15 DOWNTO 0);
  VARIABLE lab           : STRING(1 TO max_label_length);

BEGIN
  IF (pos <= len) AND NOT(str(pos) = ';') AND NOT(str(pos) = ',') THEN
    WHILE (pos <= len) AND NOT(found_comment) LOOP
      IF ((str(pos) >= '0') AND (str(pos) <= '9')) OR (str(pos) = '-') OR
         (str(pos) = '$') OR (str(pos) = '%') THEN -- Integer
        got_comma := FALSE;
        get_integer(str, len, pos, -32768, 65535, int, err);
        IF (err = 0) THEN
          IF (int < 0) THEN
            data := STD_LOGIC_VECTOR(CONV_SIGNED(int, 16));
          ELSE
            data := STD_LOGIC_VECTOR(CONV_UNSIGNED(int, 16));
          END IF;
          add_to_ram(str, ln, pos, data, "", lab_none, inst_rec);
        ELSIF (err = 1) THEN
          err_msg(str, ln, pos, "Word value missing.");
        ELSE
          err_msg(str, ln, pos, "Word value out of range (use -32768..62737 or 0..65535).");
        END IF;
      ELSIF (upcase(str(pos)) >= 'A') AND (upcase(str(pos)) <= 'Z') THEN -- Label
        got_comma := FALSE;
        get_label(str, len, ln, pos, lab);
        add_to_ram(str, ln, pos, "0000000000000000", lab, lab_abs, inst_rec);
      ELSE
        err_msg(str, ln, pos, "Word value corrupt or invalid.");
      END IF;
      IF (pos <= len) THEN
        IF (str(pos) = ' ') THEN
          swallow_spaces(str, len, pos);
          IF (pos <= len) THEN
            IF (str(pos) = ',') THEN
              pos := pos + 1;
              got_comma := TRUE;
              swallow_spaces(str, len, pos);
            ELSIF (str(pos) = ';') THEN
              found_comment := TRUE;
            ELSE
              err_msg(str, ln, pos, "Missing comma after Word value.");
            END IF;
          END IF;
        ELSIF (str(pos) = ',') THEN
          pos := pos + 1;
          got_comma := TRUE;
          swallow_spaces(str, len, pos);
        ELSE
          err_msg(str, ln, pos, "Word value corrupt or invalid.");
        END IF;
      END IF;
    END LOOP;
    IF got_comma THEN
      err_msg(str, ln, pos, "Word value missing after comma.");
    END IF;
  ELSE
    err_msg(str, ln, pos, "Word value missing after 'words' keyword.");
  END IF;
END compile_words;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE compile_bytes(str      : IN    STRING;          -- Line of source code
                        len      : IN    NATURAL;         -- Length of line
                        ln       : IN    NATURAL;         -- Line number
                        pos      : INOUT NATURAL;         -- Character position
                        inst_rec : INOUT INSTRUCTIONS) IS -- Instruction Record

  VARIABLE got_comma     : BOOLEAN := TRUE;
  VARIABLE char_msb      : BOOLEAN := TRUE;
  VARIABLE found_comment : BOOLEAN := FALSE;
  VARIABLE err           : NATURAL;
  VARIABLE int           : INTEGER;
  VARIABLE two_chars     : STD_LOGIC_VECTOR(15 DOWNTO 0);
  VARIABLE byte_count    : NATURAL := 0;
  

BEGIN
  IF (pos <= len) AND NOT(str(pos) = ';') AND NOT(str(pos) = ',') THEN
    WHILE (pos <= len) AND NOT(found_comment) LOOP
      IF (str(pos) = '"') THEN
        got_comma := FALSE;
        pos := pos + 1;
        WHILE (pos <= len) AND NOT(str(pos) = '"') LOOP
          byte_count := byte_count + 1;
          IF char_msb THEN
            two_chars := STD_LOGIC_VECTOR(CONV_UNSIGNED(CHARACTER'POS(str(pos)), 8)) & "00000000";
          ELSE
            two_chars(7 DOWNTO 0) := STD_LOGIC_VECTOR(CONV_UNSIGNED(CHARACTER'POS(str(pos)), 8));
          END IF;
          IF NOT(char_msb) THEN
            add_to_ram(str, ln, pos, two_chars, "", lab_none, inst_rec);
          END IF;
          char_msb := NOT(char_msb);
          pos := pos + 1;
        END LOOP;
        IF (pos <= len) THEN
          pos := pos + 1;
        ELSE
          err_msg(str, ln, pos, "Missing quote after character string.");
        END IF;
      ELSIF ((str(pos) >= '0') AND (str(pos) <= '9')) OR (str(pos) = '-') OR
            (str(pos) = '$') OR (str(pos) = '%') THEN -- Integer
        got_comma := FALSE;
        get_integer(str, len, pos, -128, 255, int, err);
        IF (err = 0) THEN
          byte_count := byte_count + 1;
          IF (int < 0) THEN
            IF char_msb THEN
              two_chars := STD_LOGIC_VECTOR(CONV_SIGNED(int, 8)) & "00000000";
            ELSE
              two_chars(7 DOWNTO 0) := STD_LOGIC_VECTOR(CONV_SIGNED(int, 8));
            END IF;
          ELSE
            IF char_msb THEN
              two_chars := STD_LOGIC_VECTOR(CONV_UNSIGNED(int, 8)) & "00000000";
            ELSE
              two_chars(7 DOWNTO 0) := STD_LOGIC_VECTOR(CONV_UNSIGNED(int, 8));
            END IF;
          END IF;
          IF NOT(char_msb) THEN
            add_to_ram(str, ln, pos, two_chars, "", lab_none, inst_rec);
          END IF;
          char_msb := NOT(char_msb);
        ELSIF (err = 1) THEN
          err_msg(str, ln, pos, "Byte value missing.");
        ELSE
          err_msg(str, ln, pos, "Byte value out of range (use -128..127 or 0..255).");
        END IF;
      ELSE
        err_msg(str, ln, pos, "Byte value corrupt or invalid.");
      END IF;
      IF (pos <= len) THEN
        IF (str(pos) = ' ') THEN
          swallow_spaces(str, len, pos);
          IF (pos <= len) THEN
            IF (str(pos) = ',') THEN
              pos := pos + 1;
              got_comma := TRUE;
              swallow_spaces(str, len, pos);
            ELSIF (str(pos) = ';') THEN
              found_comment := TRUE;
            ELSE
              err_msg(str, ln, pos, "Missing comma after Byte value.");
            END IF;
          END IF;
        ELSIF (str(pos) = ',') THEN
          pos := pos + 1;
          got_comma := TRUE;
          swallow_spaces(str, len, pos);
        ELSE
          err_msg(str, ln, pos, "Byte value corrupt or invalid.");
        END IF;
      END IF;
    END LOOP;
    IF got_comma THEN
      err_msg(str, ln, pos, "Byte value missing after comma.");
    END IF;
    IF (byte_count > 0) THEN
      IF NOT(char_msb) THEN
        add_to_ram(str, ln, pos, two_chars, "", lab_none, inst_rec);
        err_msg(str, ln, pos, "Uneven number of bytes found in byte sequence.");
      END IF;
    ELSE
      err_msg(str, ln, pos, "Byte value missing after 'bytes' keyword.");
    END IF;
  ELSE
    err_msg(str, ln, pos, "Byte value missing after 'bytes' keyword.");
  END IF;
END compile_bytes;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE compile_inst(str      : IN    STRING;                       -- Line of source code
                       len      : IN    NATURAL;                      -- Length of line
                       ln       : IN    NATURAL;                      -- Line number
                       pos      : INOUT NATURAL;                      -- Character position
                       ins      : IN    STD_LOGIC_VECTOR(3 DOWNTO 0); -- Instruction
                       quick    : IN    BOOLEAN;                      -- Quick Instruction
                       inst_rec : INOUT INSTRUCTIONS) IS              -- Instruction Record

  VARIABLE src_mode : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE src_reg  : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE src_data : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
  VARIABLE src_lab  : STRING(1 TO max_label_length) := (OTHERS => NUL);
  VARIABLE dst_mode : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE dst_reg  : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE dst_data : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
  VARIABLE dst_lab  : STRING(1 TO max_label_length) := (OTHERS => NUL);
  VARIABLE inst     : STD_LOGIC_VECTOR(15 DOWNTO 0);
  VARIABLE swap_ops : BOOLEAN := FALSE;

BEGIN

  --
  -- Get Source Operand.
  --
  IF quick THEN
    --
    -- Get Source Operand (#-8..7, #0..15 or #Label).
    --
    get_operand(str, len, ln, pos, "00100100000", src_mode, src_reg, src_data, src_lab);
    IF (src_lab(1) = NUL) THEN
      IF (CONV_INTEGER(UNSIGNED(src_data)) >= 0) AND (CONV_INTEGER(UNSIGNED(src_data)) <= 7) THEN
        src_mode := mode_imm_quick;
        src_reg := src_data(2 DOWNTO 0);
      ELSE
        IF (ins = inst_shl) OR (ins = inst_shr) OR (ins = inst_add) OR
           (ins = inst_sub) OR (ins = inst_mulu) OR (ins = inst_divu) THEN
          IF (CONV_INTEGER(UNSIGNED(src_data)) >= 0) AND (CONV_INTEGER(UNSIGNED(src_data)) <= 15) THEN
            src_mode := mode_imm_quick;
            src_reg := src_data(2 DOWNTO 0);
            swap_ops := TRUE;
          ELSE
            err_msg(str, ln, pos, "Instruction field out of range (use 0..15).");
          END IF;
        ELSE -- ins = Move, And, Or, Xor, Cmp, Muls or Divs
          IF (CONV_INTEGER(SIGNED(src_data)) >= -8) AND (CONV_INTEGER(SIGNED(src_data)) <= -1) THEN
            src_mode := mode_imm_quick;
            src_reg := src_data(2 DOWNTO 0);
            swap_ops := TRUE;
          ELSE
            err_msg(str, ln, pos, "Instruction field out of range (use -8..7).");
          END IF;
        END IF;
      END IF;
    ELSE
      src_mode := mode_imm_quick;
      src_reg := "000";
    END IF;
  ELSE
    --
    -- Get Source Operand (full set).
    --
    get_operand(str, len, ln, pos, "11111111111", src_mode, src_reg, src_data, src_lab);
  END IF;

  --
  -- Check for Seperation Comma between Source and Destination Operands.
  --
  swallow_spaces(str, len, pos);
  IF (pos <= len) AND (str(pos) = ',') THEN
    pos := pos + 1;
  ELSE
    err_msg(str, ln, pos, "Comma missing after instruction field.");
  END IF;
  swallow_spaces(str, len, pos);

  --
  -- Get Destination Operand.
  --
  get_operand(str, len, ln, pos, "11011011111", dst_mode, dst_reg, dst_data, dst_lab);

  --
  -- Check for Optimisations.
  --
  IF (src_lab(1) = NUL) AND (src_mode = mode_extend) AND (src_reg = mode_ext_imm) THEN -- Src Op = Immediate
    IF (ins = inst_move) AND ((dst_mode = mode_dir_lower) OR (dst_mode = mode_dir_upper)) THEN -- Dst Op = Dn/An
      IF (CONV_INTEGER(SIGNED(src_data)) >= -128) AND (CONV_INTEGER(SIGNED(src_data)) <= 127) THEN
        optim_msg(str, ln, pos, "Immediate Data in range -128 to 127, use LoadQ instruction to improve code density.");
      END IF;
    ELSE
      IF (ins = inst_shl) OR (ins = inst_shr) OR (ins = inst_add) OR
         (ins = inst_sub) OR (ins = inst_mulu) OR (ins = inst_divu) THEN
        IF (CONV_INTEGER(UNSIGNED(src_data)) >= 0) AND (CONV_INTEGER(UNSIGNED(src_data)) <= 15) THEN
          optim_msg(str, ln, pos, "Immediate Data in range 0 to 15, use Quick instruction to improve code density.");
        END IF;
      ELSE -- ins = Move, And, Or, Xor, Cmp, Muls or Divs
        IF (CONV_INTEGER(SIGNED(src_data)) >= -8) AND (CONV_INTEGER(SIGNED(src_data)) <= 7) THEN
          optim_msg(str, ln, pos, "Immediate Data in range -8 to 7, use Quick instruction to improve code density.");
        END IF;
      END IF;
    END IF;
  END IF;

  --
  -- Create Instruction Binary.
  --
  IF swap_ops THEN
    inst := ins & src_mode & src_reg & dst_mode & dst_reg;
  ELSE
    inst := ins & dst_mode & dst_reg & src_mode & src_reg;
  END IF;
  IF NOT(src_lab(1) = NUL) AND (src_mode = mode_imm_quick) THEN
    IF (ins = inst_shl) OR (ins = inst_shr) OR (ins = inst_add) OR
       (ins = inst_sub) OR (ins = inst_mulu) OR (ins = inst_divu) THEN
      add_to_ram(str, ln, pos, inst, src_lab, lab_immq2, inst_rec); -- 0 to 15
    ELSE -- ins = Move, And, Or, Xor, Cmp, Muls or Divs
      add_to_ram(str, ln, pos, inst, src_lab, lab_immq1, inst_rec); -- -8 to 7
    END IF;
  ELSE
    add_to_ram(str, ln, pos, inst, "", lab_none, inst_rec);
  END IF;

  --
  -- Create Extra Source Data Words (if any).
  --
  IF (src_mode = mode_indir_dsp) OR
     ((src_mode = mode_extend) AND (src_reg = mode_ext_abs)) OR
     ((src_mode = mode_extend) AND (src_reg = mode_ext_imm)) THEN
    IF (src_lab(1) = NUL) THEN
      add_to_ram(str, ln, pos, src_data, "", lab_none, inst_rec);
    ELSIF (src_mode = mode_indir_dsp) THEN -- d(An)
      add_to_ram(str, ln, pos, src_data, src_lab, lab_dsp, inst_rec);
    ELSIF ((src_mode = mode_extend) AND (src_reg = mode_ext_abs)) THEN -- Label
      add_to_ram(str, ln, pos, src_data, src_lab, lab_abs, inst_rec);
    ELSE -- #n
      IF (ins = inst_shl) OR (ins = inst_shr) OR (ins = inst_add) OR
         (ins = inst_sub) OR (ins = inst_mulu) OR (ins = inst_divu) THEN
        add_to_ram(str, ln, pos, src_data, src_lab, lab_imm2, inst_rec); -- Opt on 0 to 15
      ELSE -- ins = Move, And, Or, Xor, Cmp, Muls or Divs
        add_to_ram(str, ln, pos, src_data, src_lab, lab_imm1, inst_rec); -- Opt on -8 to 7
      END IF;
    END IF;
  END IF;

  --
  -- Create Extra Destination Data Words (if any).
  --
  IF (dst_mode = mode_indir_dsp) OR ((dst_mode = mode_extend) AND (dst_reg = mode_ext_abs)) THEN
    IF (dst_lab(1) = NUL) THEN -- d(An) or Label
      add_to_ram(str, ln, pos, dst_data, "", lab_none, inst_rec);
    ELSIF (dst_mode = mode_indir_dsp) THEN -- Label(An)
      add_to_ram(str, ln, pos, dst_data, dst_lab, lab_dsp, inst_rec);
    ELSE -- Label
      add_to_ram(str, ln, pos, dst_data, dst_lab, lab_abs, inst_rec);
    END IF;
  END IF;

  --
  -- Check End of Line.
  --
  check_line_end(str, len, ln, pos);

END compile_inst;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE compile_load(str      : IN    STRING;                       -- Line of source code
                       len      : IN    NATURAL;                      -- Length of line
                       ln       : IN    NATURAL;                      -- Line number
                       pos      : INOUT NATURAL;                      -- Character position
                       ins      : IN    STD_LOGIC_VECTOR(3 DOWNTO 0); -- Instruction
                       inst_rec : INOUT INSTRUCTIONS) IS              -- Instruction Record

  VARIABLE src_mode : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE src_reg  : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE src_data : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
  VARIABLE src_lab  : STRING(1 TO max_label_length) := (OTHERS => NUL);
  VARIABLE dst_mode : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE dst_reg  : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE dst_data : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
  VARIABLE dst_lab  : STRING(1 TO max_label_length) := (OTHERS => NUL);
  VARIABLE inst     : STD_LOGIC_VECTOR(15 DOWNTO 0);

BEGIN

  --
  -- Get Source Operand (#-128..127 or #Label).
  --
  get_operand(str, len, ln, pos, "00100100000", src_mode, src_reg, src_data, src_lab);
  IF (CONV_INTEGER(SIGNED(src_data)) >= -128) AND (CONV_INTEGER(SIGNED(src_data)) <= 127) THEN
    inst(7 DOWNTO 0) := src_data(7 DOWNTO 0);
  ELSE
    err_msg(str, ln, pos, "Instruction field out of range (use -128..127).");
  END IF;

  --
  -- Check for Seperation Comma between Source and Destination Operands.
  --
  swallow_spaces(str, len, pos);
  IF (pos <= len) AND (str(pos) = ',') THEN
    pos := pos + 1;
  ELSE
    err_msg(str, ln, pos, "Comma missing after instruction field.");
  END IF;
  swallow_spaces(str, len, pos);

  --
  -- Get Destination Operand (D0..D7, A0..A5, SP or PC).
  --
  get_operand(str, len, ln, pos, "00000000001", dst_mode, dst_reg, dst_data, dst_lab);

  --
  -- Create Instruction Binary.
  --
  IF (src_lab(1) = NUL) THEN
    inst := ins & dst_mode(0) & dst_reg & src_data(7 DOWNTO 0);
    add_to_ram(str, ln, pos, inst, "", lab_none, inst_rec);
  ELSE
    inst := ins & dst_mode(0) & dst_reg & "00000000";
    add_to_ram(str, ln, pos, inst, src_lab, lab_load, inst_rec);
  END IF;

  --
  -- Check End of Line.
  --
  check_line_end(str, len, ln, pos);

END compile_load;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE compile_not(str      : IN    STRING;          -- Line of source code
                      len      : IN    NATURAL;         -- Length of line
                      ln       : IN    NATURAL;         -- Line number
                      pos      : INOUT NATURAL;         -- Character position
                      inst_rec : INOUT INSTRUCTIONS) IS -- Instruction Record

  VARIABLE dst_mode : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE dst_reg  : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE dst_data : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
  VARIABLE dst_lab  : STRING(1 TO max_label_length) := (OTHERS => NUL);
  VARIABLE inst     : STD_LOGIC_VECTOR(15 DOWNTO 0);

BEGIN

  --
  -- Get Destination Operand.
  --
  get_operand(str, len, ln, pos, "11011011111", dst_mode, dst_reg, dst_data, dst_lab);

  --
  -- Create Instruction Binary.
  --
  inst := STD_LOGIC_VECTOR(CONV_UNSIGNED(inst_xor, 4)) & mode_imm_quick & "111" & dst_mode & dst_reg;
  add_to_ram(str, ln, pos, inst, "", lab_none, inst_rec);

  --
  -- Create Extra Data Words (if any).
  --
  IF (dst_mode = mode_indir_dsp) OR ((dst_mode = mode_extend) AND (dst_reg = mode_ext_abs)) THEN
    IF (dst_lab(1) = NUL) THEN -- d(An) or Label
      add_to_ram(str, ln, pos, dst_data, "", lab_none, inst_rec);
    ELSIF (dst_mode = mode_indir_dsp) THEN -- Label(An)
      add_to_ram(str, ln, pos, dst_data, dst_lab, lab_dsp, inst_rec);
    ELSE -- Label
      add_to_ram(str, ln, pos, dst_data, dst_lab, lab_abs, inst_rec);
    END IF;
  END IF;

  --
  -- Check End of Line.
  --
  check_line_end(str, len, ln, pos);

END compile_not;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE compile_rts(str      : IN    STRING;          -- Line of source code
                      len      : IN    NATURAL;         -- Length of line
                      ln       : IN    NATURAL;         -- Line number
                      pos      : INOUT NATURAL;         -- Character position
                      inst_rec : INOUT INSTRUCTIONS) IS -- Instruction Record

  VARIABLE branch_type : STD_LOGIC_VECTOR(3 DOWNTO 0);
  VARIABLE inst        : STD_LOGIC_VECTOR(15 DOWNTO 0);
  VARIABLE condition   : STRING(1 TO 3) := (OTHERS => NUL);
  VARIABLE got_cond    : BOOLEAN := TRUE;

BEGIN
  --
  -- Get Rts Condition (if any).
  --
  condition := str(pos TO pos+2);
  CASE condition IS
    WHEN "alw"  => branch_type := branch_alw;
    WHEN "nev"  => branch_type := branch_nev;
    WHEN "zrs"  => branch_type := branch_zrs;
    WHEN "zrc"  => branch_type := branch_zrc;
    WHEN "ngs"  => branch_type := branch_ngs;
    WHEN "ngc"  => branch_type := branch_ngc;
    WHEN "uos"  => branch_type := branch_uos;
    WHEN "uoc"  => branch_type := branch_uoc;
    WHEN "sos"  => branch_type := branch_sos;
    WHEN "soc"  => branch_type := branch_soc;
    WHEN "ugt"  => branch_type := branch_ugt;
    WHEN "ule"  => branch_type := branch_ule;
    WHEN "sgt"  => branch_type := branch_sgt;
    WHEN "sge"  => branch_type := branch_sge;
    WHEN "slt"  => branch_type := branch_slt;
    WHEN "sle"  => branch_type := branch_sle;
    WHEN "equ"  => branch_type := branch_equ;
    WHEN "neq"  => branch_type := branch_neq;
    WHEN "ult"  => branch_type := branch_ult;
    WHEN "uge"  => branch_type := branch_uge;
    WHEN OTHERS => branch_type := branch_alw;
                   got_cond := FALSE;
  END CASE;

  --
  -- Check End of Line.
  --
  IF NOT(str(pos) = ';') THEN
    IF got_cond THEN
      IF (str(pos+3) = ' ') OR (str(pos+3) = NUL) THEN
        pos := pos + 3;
        check_line_end(str, len, ln, pos);
      ELSE
        err_msg(str, ln, pos, "Rts condition field invalid or corrupt.");
      END IF;
    ELSE
      check_line_end(str, len, ln, pos);
    END IF;
  END IF;

  --
  -- Create Instruction Binary.
  --
  inst := STD_LOGIC_VECTOR(CONV_UNSIGNED(inst_bra, 4)) & branch_type & "10" & mode_indir_pi & "110";
  add_to_ram(str, ln, pos, inst, "", lab_none, inst_rec);

END compile_rts;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE compile_pupo(str      : IN    STRING;                       -- Line of source code
                       len      : IN    NATURAL;                      -- Length of line
                       ln       : IN    NATURAL;                      -- Line number
                       pos      : INOUT NATURAL;                      -- Character position
                       ins      : IN    STD_LOGIC_VECTOR(3 DOWNTO 0); -- Instruction
                       push     : IN    BOOLEAN;                      -- Push (or Pop)
                       lower    : IN    BOOLEAN;                      -- Lower (or Upper)
                       inst_rec : INOUT INSTRUCTIONS) IS              -- Instruction Record

  VARIABLE get_reg   : BOOLEAN := TRUE;
  VARIABLE tmp       : STRING(1 TO 2) := (OTHERS => NUL);
  VARIABLE reg_mask  : STD_LOGIC_VECTOR(7 DOWNTO 0) := (OTHERS => '0');
  VARIABLE upper_bit : STD_LOGIC;
  VARIABLE inst      : STD_LOGIC_VECTOR(15 DOWNTO 0);

BEGIN

  WHILE (get_reg) AND (pos + 1 <= len) LOOP

    --
    -- Get Register.
    --
    tmp := upcase(str(pos)) & upcase(str(pos+1));
    CASE lower IS
      WHEN TRUE =>
        CASE tmp IS
          WHEN "D0" => reg_mask := reg_mask OR "00000001";
          WHEN "D1" => reg_mask := reg_mask OR "00000010";
          WHEN "D2" => reg_mask := reg_mask OR "00000100";
          WHEN "D3" => reg_mask := reg_mask OR "00001000";
          WHEN "A0" => reg_mask := reg_mask OR "00010000";
          WHEN "A1" => reg_mask := reg_mask OR "00100000";
          WHEN "A2" => reg_mask := reg_mask OR "01000000";
          WHEN "A3" => reg_mask := reg_mask OR "10000000";
          WHEN OTHERS =>
            err_msg(str, ln, pos, "Register corrupt, use D0, D1, D2, D3, A0, A1, A2 or A3.");
        END CASE;
      WHEN FALSE =>
        CASE tmp IS
          WHEN "D4" => reg_mask := reg_mask OR "00000001";
          WHEN "D5" => reg_mask := reg_mask OR "00000010";
          WHEN "D6" => reg_mask := reg_mask OR "00000100";
          WHEN "D7" => reg_mask := reg_mask OR "00001000";
          WHEN "A4" => reg_mask := reg_mask OR "00010000";
          WHEN "A5" => reg_mask := reg_mask OR "00100000";
          WHEN "SP" => reg_mask := reg_mask OR "01000000";
          WHEN "PC" => reg_mask := reg_mask OR "10000000";
          WHEN OTHERS =>
            err_msg(str, ln, pos, "Register corrupt, use D4, D5, D6, D7, A4, A5, SP or PC.");
        END CASE;
    END CASE;
    pos := pos + 2;

    --
    -- Check for Seperation Comma between registers.
    --
    IF (pos <= len) THEN
      IF (str(pos) = ' ') THEN
        swallow_spaces(str, len, pos);
        IF (str(pos) = ',') THEN
          pos := pos + 1;
          swallow_spaces(str, len, pos);
          get_reg := TRUE;
        ELSIF (str(pos) = ';') THEN
          pos := pos - 1;
          check_line_end(str, len, ln, pos);
          get_reg := FALSE;
        ELSE
          err_msg(str, ln, pos, "Comma missing after register.");
        END IF;
      ELSIF (str(pos) = ',') THEN
        pos := pos + 1;
        swallow_spaces(str, len, pos);
        get_reg := TRUE;
      ELSE
        IF lower THEN
          err_msg(str, ln, pos, "Register corrupt, use D0, D1, D2, D3, A0, A1, A2 or A3.");
        ELSE
          err_msg(str, ln, pos, "Register corrupt, use D4, D5, D6, D7, A4, A5, SP or PC.");
        END IF;
      END IF;
    ELSE
      get_reg := FALSE;
    END IF;

  END LOOP;

  --
  -- Create Instruction Binary.
  --
  IF (get_reg) THEN
    err_msg(str, ln, pos, "Register expected.");
  ELSE
    IF lower THEN
      upper_bit := '0';
    ELSE
      upper_bit := '1';
    END IF;
    IF push THEN
      inst := '0' & upper_bit & reg_mask(7 DOWNTO 6) & mode_imm_quick & reg_mask(5 DOWNTO 3) & mode_imm_quick & reg_mask(2 DOWNTO 0);
    ELSE
      inst := '0' & upper_bit & reg_mask(0) & reg_mask(1) & mode_extend & mode_ext_imm & reg_mask(2) & reg_mask(3) & reg_mask(4) & reg_mask(5) & reg_mask(6) & reg_mask(7);
    END IF;
    add_to_ram(str, ln, pos, inst, "", lab_none, inst_rec);
  END IF;

END compile_pupo;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE compile_bra(str      : IN    STRING;                       -- Line of source code
                      len      : IN    NATURAL;                      -- Length of line
                      ln       : IN    NATURAL;                      -- Line number
                      pos      : INOUT NATURAL;                      -- Character position
                      ins      : IN    STD_LOGIC_VECTOR(3 DOWNTO 0); -- Instruction
                      quick    : IN    BOOLEAN;                      -- Quick instruction
                      inst_rec : INOUT INSTRUCTIONS) IS              -- Instruction Record

  VARIABLE mode        : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE reg         : STD_LOGIC_VECTOR(2 DOWNTO 0);
  VARIABLE data        : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
  VARIABLE lab         : STRING(1 TO max_label_length) := (OTHERS => NUL);
  VARIABLE branch_type : STD_LOGIC_VECTOR(3 DOWNTO 0);
  VARIABLE inst        : STD_LOGIC_VECTOR(15 DOWNTO 0);
  VARIABLE condition   : STRING(1 TO 3) := (OTHERS => NUL);
  VARIABLE got_cond    : BOOLEAN := TRUE;

BEGIN

  --
  -- Get Branch Condition (if any).
  --
  condition := str(pos TO pos+2);
  CASE condition IS
    WHEN "alw"  => branch_type := branch_alw;
    WHEN "nev"  => branch_type := branch_nev;
    WHEN "zrs"  => branch_type := branch_zrs;
    WHEN "zrc"  => branch_type := branch_zrc;
    WHEN "ngs"  => branch_type := branch_ngs;
    WHEN "ngc"  => branch_type := branch_ngc;
    WHEN "uos"  => branch_type := branch_uos;
    WHEN "uoc"  => branch_type := branch_uoc;
    WHEN "sos"  => branch_type := branch_sos;
    WHEN "soc"  => branch_type := branch_soc;
    WHEN "ugt"  => branch_type := branch_ugt;
    WHEN "ule"  => branch_type := branch_ule;
    WHEN "sgt"  => branch_type := branch_sgt;
    WHEN "sge"  => branch_type := branch_sge;
    WHEN "slt"  => branch_type := branch_slt;
    WHEN "sle"  => branch_type := branch_sle;
    WHEN "equ"  => branch_type := branch_equ;
    WHEN "neq"  => branch_type := branch_neq;
    WHEN "ult"  => branch_type := branch_ult;
    WHEN "uge"  => branch_type := branch_uge;
    WHEN OTHERS => branch_type := branch_alw;
                   got_cond := FALSE;
  END CASE;

  --
  -- Check for Seperation Comma between Source and Destination Operands (if any).
  --
  IF got_cond AND ((str(pos+3) = ' ') OR (str(pos+3) = ',')) THEN -- Got Condition
    pos := pos + 3;
    swallow_spaces(str, len, pos);
    IF (pos <= len) AND (str(pos) = ',') THEN
      pos := pos + 1;
    ELSE
      err_msg(str, ln, pos, "Comma missing after instruction field.");
    END IF;
    swallow_spaces(str, len, pos);
  END IF;

  --
  -- Get Destination Operand.
  --
  IF (quick) THEN
    get_operand(str, len, ln, pos, "00110100000", mode, reg, data, lab);
    IF (lab(1) = NUL) THEN -- ImmQ (Valid range = 0..7 only)
      IF (CONV_INTEGER(UNSIGNED(data)) >= 0) AND (CONV_INTEGER(UNSIGNED(data)) <= 7) THEN
        inst := ins & branch_type & "10" & mode_imm_quick & data(2 DOWNTO 0);
        add_to_ram(str, ln, pos, inst, lab, lab_none, inst_rec);
      ELSE
        err_msg(str, ln, pos, "Branch (Quick) immediate out of range (use 0..7).");
      END IF;
    ELSE
      IF (reg = mode_ext_imm) THEN -- ImmQ #Label (valid range = 0..7 only)
        inst := ins & branch_type & "10" & mode_imm_quick & "000";
        add_to_ram(str, ln, pos, inst, lab, lab_braqimm, inst_rec);
      ELSE -- ImmQ Disp (valid range = -64 to 127 only)
        inst := ins & branch_type & "00000000";
        add_to_ram(str, ln, pos, inst, lab, lab_braq, inst_rec);
      END IF;
    END IF;
  ELSE
    get_operand(str, len, ln, pos, "01110111111", mode, reg, data, lab);
    IF (lab(1) = NUL) THEN
      inst := ins & branch_type & "10" & mode & reg;
      add_to_ram(str, ln, pos, inst, "", lab_none, inst_rec);
      IF (mode = mode_indir_dsp) OR ((mode = mode_extend) AND (reg = mode_ext_imm)) THEN
        add_to_ram(str, ln, pos, data, "", lab_none, inst_rec);
      END IF;
      IF (mode = mode_extend) AND (reg = mode_ext_imm) THEN
        IF (CONV_INTEGER(UNSIGNED(data)) >= 0) AND (CONV_INTEGER(UNSIGNED(data)) <= 7) THEN
          optim_msg(str, ln, pos, "Immediate Data in range 0 to 7, use Quick instruction to improve code density.");
        END IF;
      END IF;
    ELSE -- #Lab, Lab or Lab(An)
      inst := ins & branch_type & "10"  & mode & reg;
      add_to_ram(str, ln, pos, inst, "", lab_none, inst_rec);
      IF (mode = mode_indir_dsp) THEN
        add_to_ram(str, ln, pos, data, lab, lab_dsp, inst_rec);
      ELSIF (reg = mode_ext_imm) THEN
        add_to_ram(str, ln, pos, data, lab, lab_imm3, inst_rec);
      ELSE
        add_to_ram(str, ln, pos, data, lab, lab_bra, inst_rec);
      END IF;
    END IF;
  END IF;

  --
  -- Check End of Line.
  --
  check_line_end(str, len, ln, pos);

END compile_bra;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

PROCEDURE add_label(str       : IN    STRING;                        -- Line of source code
                    len       : IN    NATURAL;                       -- Length of line
                    ln        : IN    NATURAL;                       -- Line number
                    pos       : INOUT NATURAL;                       -- Character position
                    lab       : IN    STRING(1 TO max_label_length); -- Label
                    value     : IN    INTEGER;                       -- Label Value
                    label_rec : INOUT LABELS) IS                     -- Instruction Record

  VARIABLE index : NATURAL;

BEGIN
  IF (label_rec.count < max_label_entries) THEN
    FOR index IN 1 TO label_rec.count LOOP
      IF (label_rec.list(index).name = lab) THEN
        err_msg(str, ln, pos, "Label already in use.");
      END IF;
    END LOOP;
    label_rec.count := label_rec.count + 1;
    label_rec.list(label_rec.count).name := lab;
    label_rec.list(label_rec.count).value := value;
  ELSE
    err_msg(str, ln, pos, "Label table full, reduce number of labels or increase table size.");
  END IF;
END add_label;

--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--

BEGIN

  do_read : PROCESS

    FILE     fin         : TEXT OPEN READ_MODE IS assembly_file;
    FILE     coe_fout    : TEXT OPEN WRITE_MODE IS coregen_path & coe_file;
    FILE     xco_fout    : TEXT OPEN WRITE_MODE IS coregen_path & xco_file;
    FILE     sp2_fout    : TEXT OPEN WRITE_MODE IS source_path & sp2_file;
    FILE     sp3_fout    : TEXT OPEN WRITE_MODE IS source_path & sp3_file;

    VARIABLE fline       : LINE;
    VARIABLE fchar       : CHARACTER;
    VARIABLE fgood       : BOOLEAN;
    VARIABLE str         : STRING(1 TO max_code_length) := (OTHERS => NUL);
    VARIABLE len         : NATURAL;
    VARIABLE line_out    : LINE;
    VARIABLE ln          : NATURAL := 1;
    VARIABLE lab         : STRING(1 TO max_label_length);
    VARIABLE lab_len     : NATURAL;
    VARIABLE pos         : NATURAL;
    VARIABLE inst        : NATURAL;
    VARIABLE inst_i      : STD_LOGIC_VECTOR(3 DOWNTO 0);
    VARIABLE index       : NATURAL;
    VARIABLE index2      : NATURAL;
    VARIABLE index3      : NATURAL;
    VARIABLE inst_rec    : INSTRUCTIONS;
    VARIABLE label_rec   : LABELS;
    VARIABLE found_label : BOOLEAN;
    VARIABLE tmp         : STD_LOGIC_VECTOR(3 DOWNTO 0);
    VARIABLE source      : CODE_LINES;
    VARIABLE vec         : STD_LOGIC_VECTOR(3 DOWNTO 0);
    VARIABLE addr_msb    : NATURAL;
    VARIABLE data_msb    : NATURAL;
    VARIABLE one_more    : NATURAL;

  BEGIN

    IF NOT((ram_size = 256) OR (ram_size = 512) OR (ram_size = 1024) OR
           (ram_size = 2048) OR (ram_size = 4096) OR (ram_size = 8192) OR
           (ram_size = 16384) OR (ram_size = 32768) OR (ram_size = 63336)) THEN
      WRITE(line_out, STRING'("**"));
      WRITELINE(OUTPUT, line_out);
      WRITE(line_out, STRING'("** Invalid RAM_SIZE constant in assembler.vhd, use either 256, 512, 1024, 2048, 4096, 8192, 16384, 32768 or 63336."));
      WRITELINE(OUTPUT, line_out);
      WRITE(line_out, STRING'("**"));
      WRITELINE(OUTPUT, line_out);
      ASSERT FALSE SEVERITY FAILURE;
    END IF;


    --
    -- Clear Instruction Record.
    --
    FOR index IN 0 TO ram_size-1 LOOP
      inst_rec.list(index).inst    := STD_LOGIC_VECTOR(CONV_UNSIGNED(inst_bra, 4)) & branch_alw & "11" & "111111"; -- Label BraQ alw,Label
      inst_rec.list(index).type_1  := 0;
      inst_rec.list(index).label_1 := (OTHERS => NUL);
      inst_rec.list(index).line    := 0;
    END LOOP;
    inst_rec.ptr := 0;

    --
    -- Clear Label Record.
    --
    FOR index IN 1 TO max_label_entries LOOP
      label_rec.list(index).name  := (OTHERS => NUL);
      label_rec.list(index).value := 0;
    END LOOP;
    label_rec.count := 0;

    --
    -- Display program information.
    --
    WRITELINE(OUTPUT, line_out);
    WRITE(line_out, STRING'("Proc16 Micro Assembler - Copyright (c) 2004 Steve Haywood (steve.haywood@ukonline.co.uk)"));
    WRITELINE(OUTPUT, line_out);
    WRITELINE(OUTPUT, line_out);
    WRITE(line_out, STRING'("Assembling : "));
    WRITE(line_out, assembly_file);
    WRITE(line_out, STRING'(" (Pass 1)."));
    WRITELINE(OUTPUT, line_out);
    WRITELINE(OUTPUT, line_out);

    IF DEBUG_MSG THEN
      IF (inst_rec.ptr = 0) THEN
        WRITE(line_out, STRING'("Word - Binary           - Line - Source"));
        WRITELINE(OUTPUT, line_out);
        WRITELINE(OUTPUT, line_out);
      END IF;
    END IF;

    WHILE NOT(ENDFILE(fin)) LOOP

      --
      -- Read Line.
      --
      READLINE(fin, fline);

      --
      -- Copy line into buffer.
      --
      len := 0;
      READ(fline, fchar, fgood);
      WHILE fgood AND (len <= max_code_length) LOOP
        IF (len < max_code_length) THEN
          len := len + 1;
          str(len) := fchar;
          READ(fline, fchar, fgood);
        ELSE
          err_msg(str, ln, 0, "Source line buffer full, reduce line length or increase buffer size.");
        END IF;
      END LOOP;
      IF (ln <= max_code_lines) THEN
        source(ln) := str;
      ELSE
        err_msg(str, ln, 0, "Source code buffer full, reduce lines of code or increase buffer size.");
      END IF;


      IF (len > 0) AND NOT(str(1) = ';') THEN

        --
        -- Reset character position to start of line.
        --
        pos := 1;

        --
        -- Read and add line label if present.
        --
        IF (str(pos) = ' ') THEN
          swallow_spaces(str, len, pos);
        ELSE
          get_label(str, len, ln, pos, lab);
          IF (pos <= len) THEN
            IF (str(pos) = ' ') THEN
              add_label(str, len, ln, pos, lab, inst_rec.ptr, label_rec);
              swallow_spaces(str, len, pos);
            ELSE
              err_msg(str, ln, pos, "Label invalid (use characters a-z and A-Z only).");
            END IF;
          ELSE
            err_msg(str, ln, pos, "Label without data or instruction to follow.");
          END IF;
        END IF;

        IF NOT(pos > len) AND NOT(str(pos) = ';') THEN

          --
          -- Read instruction or keyword.
          --
          get_instruction(str, len, ln, pos, inst);
          inst_i := STD_LOGIC_VECTOR(CONV_UNSIGNED(inst, 4));

          --
          -- Compile instruction or keyword.
          --
          CASE inst IS

            WHEN inst_move | inst_and | inst_or | inst_xor | inst_shl | inst_shr |
                 inst_add | inst_sub | inst_cmp => 

              compile_inst(str, len, ln, pos, inst_i, FALSE, inst_rec);

            WHEN inst_moveq | inst_andq | inst_orq | inst_xorq | inst_shlq | inst_shrq |
                 inst_addq | inst_subq | inst_cmpq =>

              compile_inst(str, len, ln, pos, inst_i, TRUE, inst_rec);

            WHEN inst_bra | inst_bsr =>

              compile_bra(str, len, ln, pos, inst_i, FALSE, inst_rec);

            WHEN inst_braq | inst_bsrq =>

              compile_bra(str, len, ln, pos, inst_i, TRUE, inst_rec);

            WHEN inst_popu =>

              compile_pupo(str, len, ln, pos, inst_i, FALSE, FALSE, inst_rec);

            WHEN inst_popl =>

              compile_pupo(str, len, ln, pos, inst_i, FALSE, TRUE, inst_rec);

            WHEN inst_pushu =>

              compile_pupo(str, len, ln, pos, inst_i, TRUE, FALSE, inst_rec);

            WHEN inst_pushl =>

              compile_pupo(str, len, ln, pos, inst_i, TRUE, TRUE, inst_rec);

            WHEN inst_not => -- Not <dst>

              compile_not(str, len, ln, pos, inst_rec);

            WHEN inst_rts =>

              compile_rts(str, len, ln, pos, inst_rec);

            WHEN inst_nop => -- Label BraQ nev,Label

              IF (pos < len) THEN
                pos := pos - 1;
                check_line_end(str, len, ln, pos);
              END IF;
              add_to_ram(str, ln, pos, STD_LOGIC_VECTOR(CONV_UNSIGNED(inst_bra, 4)) & branch_nev & "11" & "111111", lab, lab_none, inst_rec);

            WHEN inst_rti =>

              IF (pos < len) THEN
                pos := pos - 1;
                check_line_end(str, len, ln, pos);
              END IF;
              add_to_ram(str, ln, pos, STD_LOGIC_VECTOR(CONV_UNSIGNED(inst_bra, 4)) & branch_alw & "10" & mode_extend & "101", lab, lab_none, inst_rec);

            WHEN inst_load => 

              compile_load(str, len, ln, pos, inst_i, inst_rec);

            WHEN inst_bytes =>

              compile_bytes(str, len, ln, pos, inst_rec);

            WHEN inst_words =>

              compile_words(str, len, ln, pos, inst_rec);

            WHEN OTHERS =>

              err_msg(str, ln, pos, "Unknown instruction.");

          END CASE;

        ELSIF DEBUG_MSG THEN
          WRITE(line_out, STRING'("     -                  - "));
          WRITE(line_out, ln, right, 4);
          WRITE(line_out, STRING'(" - "));
          WRITE(line_out, str);
          WRITELINE(OUTPUT, line_out);
        END IF;
      ELSIF DEBUG_MSG THEN
        WRITE(line_out, STRING'("     -                  - "));
        WRITE(line_out, ln, right, 4);
        WRITE(line_out, STRING'(" - "));
        WRITE(line_out, str);
        WRITELINE(OUTPUT, line_out);
      END IF;

      --
      -- Clear Line & Label.
      --
      str := (OTHERS => NUL);
      lab := (OTHERS => NUL);

      --
      -- Increment Line Number.
      --
      ln := ln + 1;

    END LOOP;

    IF DEBUG_MSG THEN
      WRITELINE(OUTPUT, line_out);
    END IF;

    WRITE(line_out, STRING'("Assembling : "));
    WRITE(line_out, assembly_file);
    WRITE(line_out, STRING'(" (Pass 2)."));
    WRITELINE(OUTPUT, line_out);
    WRITELINE(OUTPUT, line_out);

    IF DEBUG_MSG THEN
      WRITE(line_out, STRING'("Word - Label"));
      FOR index3 IN 1 TO max_label_length-4 LOOP
        WRITE(line_out, STRING'(" "));
      END LOOP;
      WRITE(line_out, STRING'("- Binary           - Line - Source"));
      WRITELINE(OUTPUT, line_out);
      WRITELINE(OUTPUT, line_out);
    END IF;

    FOR index IN 0 TO inst_rec.ptr-1 LOOP

      IF NOT(inst_rec.list(index).label_1(1) = NUL) THEN -- Find Label
        index2 := 1;
        found_label := FALSE;
        WHILE (index2 <= label_rec.count) AND NOT(found_label) LOOP
          IF (label_rec.list(index2).name = inst_rec.list(index).label_1) THEN
            found_label := TRUE;
          ELSE
            index2 := index2 + 1;
          END IF;
        END LOOP;

        IF found_label THEN

          CASE inst_rec.list(index).type_1 IS

            WHEN lab_load =>

              IF (label_rec.list(index2).value >= -128) AND (label_rec.list(index2).value <= 127) THEN
                inst_rec.list(index).inst(7 DOWNTO 0) := STD_LOGIC_VECTOR(CONV_SIGNED(label_rec.list(index2).value, 8));
              ELSE
                err_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Load (Quick) value out of range (use -128..127).");
              END IF;

            WHEN lab_imm1 =>
              IF (inst_rec.list(index-1).inst(15 DOWNTO 12) = inst_move) AND ((inst_rec.list(index-1).inst(11 DOWNTO 9) = mode_dir_lower) OR (inst_rec.list(index-1).inst(11 DOWNTO 9) = mode_dir_upper)) THEN -- Dst Op = Dn/An
                IF (label_rec.list(index2).value >= -128) AND (label_rec.list(index2).value <= 127) THEN
                  optim_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "2- Immediate Data in range -128 to 127, use LoadQ instruction to improve code density.");
                END IF;
              ELSIF (label_rec.list(index2).value >= -8) AND (label_rec.list(index2).value <= 7) THEN
                optim_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Immediate Data in range -8 to 7, use Quick instruction to improve code density.");
              END IF;
              IF (label_rec.list(index2).value < 0) THEN
                inst_rec.list(index).inst := STD_LOGIC_VECTOR(CONV_SIGNED(label_rec.list(index2).value, 16));
              ELSE
                inst_rec.list(index).inst := STD_LOGIC_VECTOR(CONV_UNSIGNED(label_rec.list(index2).value, 16));
              END IF;

            WHEN lab_imm2 =>
              IF (label_rec.list(index2).value >= 0) AND (label_rec.list(index2).value <= 15) THEN
                optim_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Immediate Data in range 0 to 15, use Quick instruction to improve code density.");
              END IF;
              IF (label_rec.list(index2).value < 0) THEN
                inst_rec.list(index).inst := STD_LOGIC_VECTOR(CONV_SIGNED(label_rec.list(index2).value, 16));
              ELSE
                inst_rec.list(index).inst := STD_LOGIC_VECTOR(CONV_UNSIGNED(label_rec.list(index2).value, 16));
              END IF;


            WHEN lab_imm3 =>
              IF (label_rec.list(index2).value >= 0) AND (label_rec.list(index2).value <= 7) THEN
                optim_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Immediate Data in range 0 to 7, use Quick instruction to improve code density.");
              END IF;
              inst_rec.list(index).inst := STD_LOGIC_VECTOR(CONV_UNSIGNED(label_rec.list(index2).value, 16));

            WHEN lab_braqimm =>
              IF (label_rec.list(index2).value >= 0) AND
                 (label_rec.list(index2).value <= 7) THEN
                inst_rec.list(index).inst(2 DOWNTO 0) :=
                  STD_LOGIC_VECTOR(CONV_UNSIGNED(label_rec.list(index2).value, 3));
              ELSE
                err_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Branch (Quick) immediate out of range (use 0..7).");
              END IF;

            WHEN lab_abs =>
              inst_rec.list(index).inst := STD_LOGIC_VECTOR(CONV_UNSIGNED(label_rec.list(index2).value, 16));

            WHEN lab_braq =>
              IF (label_rec.list(index2).value - index - 1 >= -64) AND (label_rec.list(index2).value - index - 1 <= -1) THEN
                inst_rec.list(index).inst(7 DOWNTO 0) := '1' & STD_LOGIC_VECTOR(CONV_SIGNED(label_rec.list(index2).value - index - 1, 7));
              ELSIF (label_rec.list(index2).value - index - 1 >= 0) AND (label_rec.list(index2).value - index - 1 <= 127) THEN
                inst_rec.list(index).inst(7 DOWNTO 0) := '0' & STD_LOGIC_VECTOR(CONV_UNSIGNED(label_rec.list(index2).value - index - 1, 7));
              ELSE
                err_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Branch (Quick) displacement out of range (use -64..127).");
              END IF;

            WHEN lab_bra =>
              IF (label_rec.list(index2).value - index - 1 >= -32768) AND
                 (label_rec.list(index2).value - index - 1 <= 32767) THEN
                IF (label_rec.list(index2).value - index - 1 >= -64) AND
                   (label_rec.list(index2).value - index - 1 <= 127) THEN
                  optim_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Branch Displacement in range -64 to 127, use Quick instruction to improve code density.");
                END IF;
                inst_rec.list(index).inst := STD_LOGIC_VECTOR(CONV_SIGNED(label_rec.list(index2).value - index - 1, 16));
              ELSE
                err_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Branch displacement out of range (use -32768..32767).");
              END IF;

            WHEN lab_dsp =>
              IF (label_rec.list(index2).value - index - 1 >= -32768) AND
                 (label_rec.list(index2).value - index - 1 <= 32767) THEN
                inst_rec.list(index).inst := STD_LOGIC_VECTOR(CONV_SIGNED(label_rec.list(index2).value - index - 1, 16));
              ELSE
                err_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Displacement out of range (use -32768..32767).");
              END IF;

            WHEN lab_immq1 =>
              IF (label_rec.list(index2).value >= 0) AND (label_rec.list(index2).value <= 7) THEN
                inst_rec.list(index).inst(2 DOWNTO 0) := STD_LOGIC_VECTOR(CONV_UNSIGNED(label_rec.list(index2).value, 3));
              ELSIF (label_rec.list(index2).value >= -8) AND (label_rec.list(index2).value <= -1) THEN
                tmp := STD_LOGIC_VECTOR(CONV_SIGNED(label_rec.list(index2).value, 4));
                inst_rec.list(index).inst(2 DOWNTO 0) := tmp(2 DOWNTO 0);
                inst_rec.list(index).inst(11 DOWNTO 0) := inst_rec.list(index).inst(5 DOWNTO 0) & inst_rec.list(index).inst(11 DOWNTO 6);
              ELSE
                err_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Immediate (Quick) value out of range (use -8..7).");
              END IF;

            WHEN lab_immq2 =>
              IF (label_rec.list(index2).value >= 0) AND (label_rec.list(index2).value <= 7) THEN
                inst_rec.list(index).inst(2 DOWNTO 0) := STD_LOGIC_VECTOR(CONV_UNSIGNED(label_rec.list(index2).value, 3));
              ELSIF (label_rec.list(index2).value >= 8) AND (label_rec.list(index2).value <= 15) THEN
                tmp := STD_LOGIC_VECTOR(CONV_UNSIGNED(label_rec.list(index2).value, 4));
                inst_rec.list(index).inst(2 DOWNTO 0) := tmp(2 DOWNTO 0);
                inst_rec.list(index).inst(11 DOWNTO 0) := inst_rec.list(index).inst(5 DOWNTO 0) & inst_rec.list(index).inst(11 DOWNTO 6);
              ELSE
                err_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Immediate (Quick) value out of range (use 0..15).");
              END IF;

            WHEN OTHERS =>
              NULL;
          END CASE;
        ELSE
          err_msg(source(inst_rec.list(index).line), inst_rec.list(index).line, 0, "Label not defined in source.");
        END IF;

      END IF;

      IF DEBUG_MSG THEN
        WRITE(line_out, index, right, 4);
        WRITE(line_out, STRING'(" - "));
        --
        -- Find Label associated with line (if any).
        --
        index2 := 1;
        found_label := FALSE;
        WHILE (index2 <= label_rec.count) AND NOT(found_label) LOOP
          IF (index = label_rec.list(index2).value) THEN
            found_label := TRUE;
          ELSE
            index2 := index2 + 1;
          END IF;
        END LOOP;
        --
        -- Display label (if any) and pad with spaces.
        --
        FOR index3 IN 1 TO max_label_length LOOP
          IF NOT(found_label) OR (label_rec.list(index2).name(index3) = NUL) THEN
            WRITE(line_out, STRING'(" "));
          ELSE
            WRITE(line_out, label_rec.list(index2).name(index3));
          END IF;
        END LOOP;

        WRITE(line_out, STRING'(" - "));
        WRITE(line_out, TO_BITVECTOR(inst_rec.list(index).inst));
        WRITE(line_out, STRING'(" - "));
        WRITE(line_out, inst_rec.list(index).line, right, 4);
        --
        -- Display source line.
        --
        IF (index = 0) THEN
          WRITE(line_out, STRING'(" - "));
          WRITE(line_out, source(inst_rec.list(index).line));
        ELSIF (inst_rec.list(index-1).line = inst_rec.list(index).line) THEN
          WRITE(line_out, STRING'(" + "));
        ELSE
          WRITE(line_out, STRING'(" - "));
          WRITE(line_out, source(inst_rec.list(index).line));
        END IF;

        WRITELINE(OUTPUT, line_out);
      END IF;

    END LOOP;

    IF DEBUG_MSG THEN
      WRITELINE(OUTPUT, line_out);
    END IF;

    WRITE(line_out, STRING'("Assembled "));
    WRITE(line_out, ln);
    WRITE(line_out, STRING'(" lines of code into "));
    WRITE(line_out, inst_rec.ptr);
    WRITE(line_out, STRING'(" words ("));
    WRITE(line_out, inst_rec.ptr*2);
    WRITE(line_out, STRING'(" bytes)."));
    WRITELINE(OUTPUT, line_out);
    WRITELINE(OUTPUT, line_out);

    --
    -- Write out Spartan 2/2E VHDL RAM Implementaton.
    --
    IF (ram_size = 256) OR (ram_size = 512) OR (ram_size = 1024) OR
       (ram_size = 2048) OR (ram_size = 4096) THEN

      CASE ram_size IS
        WHEN 256  => addr_msb := 7;  data_msb := 15;
        WHEN 512  => addr_msb := 8;  data_msb := 7;
        WHEN 1024 => addr_msb := 9;  data_msb := 3;
        WHEN 2048 => addr_msb := 10; data_msb := 1;
        WHEN 4096 => addr_msb := 11; data_msb := 0;
        WHEN OTHERS => NULL;
      END CASE;

      WRITE(line_out, STRING'("Creating pre-initialised Block RAM VHDL file "));
      WRITE(line_out, source_path);
      WRITE(line_out, sp2_file);
      WRITE(line_out, STRING'(" for use as a "));
      WRITE(line_out, ram_size);
      WRITE(line_out, STRING'(" word Spartan 2/2E FPGA RAM (WebPack)."));
      WRITELINE(OUTPUT, line_out);
      WRITELINE(OUTPUT, line_out);

      WRITE(line_out, STRING'("LIBRARY IEEE;"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("USE IEEE.STD_LOGIC_1164.ALL;"));
      WRITELINE(sp2_fout, line_out);
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("ENTITY memory IS"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("  PORT(clk  : IN  STD_LOGIC;"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       en   : IN  STD_LOGIC;"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       we   : IN  STD_LOGIC;"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       addr : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       din  : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       dout : OUT STD_LOGIC_VECTOR(15 DOWNTO 0)"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("      );"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("END memory;"));
      WRITELINE(sp2_fout, line_out);
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("ARCHITECTURE rtl OF memory IS"));
      WRITELINE(sp2_fout, line_out);
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("COMPONENT RAMB4_S"));
      WRITE(line_out, data_msb + 1);
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("  PORT(rst  : IN  STD_LOGIC;"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       clk  : IN  STD_LOGIC;"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       en   : IN  STD_LOGIC;"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       we   : IN  STD_LOGIC;"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       addr : IN  STD_LOGIC_VECTOR("));
      WRITE(line_out, addr_msb);
      WRITE(line_out, STRING'(" DOWNTO 0);"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       di   : IN  STD_LOGIC_VECTOR("));
      WRITE(line_out, data_msb);
      WRITE(line_out, STRING'(" DOWNTO 0);"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("       do   : OUT STD_LOGIC_VECTOR("));
      WRITE(line_out, data_msb);
      WRITE(line_out, STRING'(" DOWNTO 0)"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("      );"));
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("END COMPONENT;"));
      WRITELINE(sp2_fout, line_out);
      WRITELINE(sp2_fout, line_out);
      FOR index IN 15 DOWNTO 0 LOOP
        WRITE(line_out, STRING'("ATTRIBUTE INIT_0"));
        WRITE(line_out, vec2char(STD_LOGIC_VECTOR(CONV_SIGNED(index, 4))));
        WRITE(line_out, STRING'(" : STRING;"));
        WRITELINE(sp2_fout, line_out);
      END LOOP;
      WRITELINE(sp2_fout, line_out);
      FOR index2 IN (16 / (data_msb + 1)) - 1 DOWNTO 0 LOOP
        FOR index IN 4095 DOWNTO 0 LOOP
          IF ((index+1) MOD 256 = 0) THEN
            WRITE(line_out, STRING'("ATTRIBUTE INIT_0"));
            WRITE(line_out, vec2char(STD_LOGIC_VECTOR(CONV_SIGNED(index / 256, 4))));
            WRITE(line_out, STRING'(" OF U"));
            WRITE(line_out, index2);
            WRITE(line_out, STRING'(" : LABEL IS """));
          END IF;
          vec(index MOD 4) := inst_rec.list(index / (data_msb + 1)).inst(index MOD (data_msb + 1) + index2 * (data_msb + 1));
          IF (index MOD 4 = 0) THEN
            WRITE(line_out, vec2char(vec));
          END IF;
          IF (index MOD 256 = 0) THEN
            WRITE(line_out, STRING'(""";"));
            WRITELINE(sp2_fout, line_out);
          END IF;
        END LOOP;
        WRITELINE(sp2_fout, line_out);
      END LOOP;
      WRITE(line_out, STRING'("SIGNAL rst : STD_LOGIC;"));
      WRITELINE(sp2_fout, line_out);
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("BEGIN"));
      WRITELINE(sp2_fout, line_out);
      WRITELINE(sp2_fout, line_out);
      WRITE(line_out, STRING'("  rst <= '0';"));
      WRITELINE(sp2_fout, line_out);
      WRITELINE(sp2_fout, line_out);
      FOR index IN (16 / (data_msb + 1)) - 1 DOWNTO 0 LOOP
        WRITE(line_out, STRING'("  U"));
        WRITE(line_out, index);
        WRITE(line_out, STRING'(" : RAMB4_S"));
        WRITE(line_out, data_msb + 1);
        WRITELINE(sp2_fout, line_out);
        WRITE(line_out, STRING'("    PORT MAP"));
        WRITELINE(sp2_fout, line_out);
        WRITE(line_out, STRING'("      (rst  => rst,"));
        WRITELINE(sp2_fout, line_out);
        WRITE(line_out, STRING'("       clk  => clk,"));
        WRITELINE(sp2_fout, line_out);
        WRITE(line_out, STRING'("       en   => en,"));
        WRITELINE(sp2_fout, line_out);
        WRITE(line_out, STRING'("       we   => we,"));
        WRITELINE(sp2_fout, line_out);
        WRITE(line_out, STRING'("       addr => addr("));
        WRITE(line_out, addr_msb);
        WRITE(line_out, STRING'(" DOWNTO 0),"));
        WRITELINE(sp2_fout, line_out);
        WRITE(line_out, STRING'("       di   => din("));
        WRITE(line_out, (data_msb + 1)*(index+1)-1 );
        WRITE(line_out, STRING'(" DOWNTO "));
        WRITE(line_out, (data_msb + 1)*index);
        WRITE(line_out, STRING'("),"));
        WRITELINE(sp2_fout, line_out);
        WRITE(line_out, STRING'("       do   => dout("));
        WRITE(line_out, (data_msb + 1)*(index+1)-1 );
        WRITE(line_out, STRING'(" DOWNTO "));
        WRITE(line_out, (data_msb + 1)*index);
        WRITE(line_out, STRING'(")"));
        WRITELINE(sp2_fout, line_out);
        WRITE(line_out, STRING'("      );"));
        WRITELINE(sp2_fout, line_out);
        WRITELINE(sp2_fout, line_out);
      END LOOP;
      WRITE(line_out, STRING'("END rtl;"));
      WRITELINE(sp2_fout, line_out);

    ELSE

      WRITE(line_out, STRING'("NOT creating pre-initialised Block RAM VHDL file "));
      WRITE(line_out, source_path);
      WRITE(line_out, sp2_file);
      WRITE(line_out, STRING'(" for use as a "));
      WRITE(line_out, ram_size);
      WRITE(line_out, STRING'(" word Spartan 2/2E FPGA RAM (WebPack) - RAM TOO BIG."));
      WRITELINE(OUTPUT, line_out);
      WRITELINE(OUTPUT, line_out);

    END IF;

    --
    -- Write out Spartan 3 VHDL RAM Implementaton.
    --
    IF (ram_size = 1024) OR (ram_size = 2048) OR (ram_size = 4096) OR
       (ram_size = 8192) OR (ram_size = 16384)THEN

      CASE ram_size IS
        WHEN 1024  => addr_msb := 9;  data_msb := 15; one_more := 2;
        WHEN 2048  => addr_msb := 10; data_msb := 7;  one_more := 1;
        WHEN 4096  => addr_msb := 11; data_msb := 3;  one_more := 0;
        WHEN 8192  => addr_msb := 12; data_msb := 1;  one_more := 0;
        WHEN 16384 => addr_msb := 13; data_msb := 0;  one_more := 0;
        WHEN OTHERS => NULL;
      END CASE;

      WRITE(line_out, STRING'("Creating pre-initialised Block RAM VHDL file "));
      WRITE(line_out, source_path);
      WRITE(line_out, sp3_file);
      WRITE(line_out, STRING'(" for use as a "));
      WRITE(line_out, ram_size);
      WRITE(line_out, STRING'(" word Spartan 3 FPGA RAM (WebPack)."));
      WRITELINE(OUTPUT, line_out);
      WRITELINE(OUTPUT, line_out);
      WRITE(line_out, STRING'("LIBRARY IEEE;"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("USE IEEE.STD_LOGIC_1164.ALL;"));
      WRITELINE(sp3_fout, line_out);
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("ENTITY memory IS"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("  PORT(clk  : IN  STD_LOGIC;"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       en   : IN  STD_LOGIC;"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       we   : IN  STD_LOGIC;"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       addr : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       din  : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       dout : OUT STD_LOGIC_VECTOR(15 DOWNTO 0)"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("      );"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("END memory;"));
      WRITELINE(sp3_fout, line_out);
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("ARCHITECTURE rtl OF memory IS"));
      WRITELINE(sp3_fout, line_out);
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("COMPONENT RAMB16_S"));
      WRITE(line_out, data_msb + 1 + one_more);
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("  PORT(ssr  : IN  STD_LOGIC;"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       clk  : IN  STD_LOGIC;"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       en   : IN  STD_LOGIC;"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       we   : IN  STD_LOGIC;"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       addr : IN  STD_LOGIC_VECTOR("));
      WRITE(line_out, addr_msb);
      WRITE(line_out, STRING'(" DOWNTO 0);"));
      WRITELINE(sp3_fout, line_out);
      IF (one_more /= 0) THEN
        WRITE(line_out, STRING'("       dip  : IN  STD_LOGIC_VECTOR("));
        WRITE(line_out, one_more-1);
        WRITE(line_out, STRING'(" DOWNTO 0);"));
        WRITELINE(sp3_fout, line_out);
        WRITE(line_out, STRING'("       dop  : OUT STD_LOGIC_VECTOR("));
        WRITE(line_out, one_more-1);
        WRITE(line_out, STRING'(" DOWNTO 0);"));
        WRITELINE(sp3_fout, line_out);
      END IF;
      WRITE(line_out, STRING'("       di   : IN  STD_LOGIC_VECTOR("));
      WRITE(line_out, data_msb);
      WRITE(line_out, STRING'(" DOWNTO 0);"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("       do   : OUT STD_LOGIC_VECTOR("));
      WRITE(line_out, data_msb);
      WRITE(line_out, STRING'(" DOWNTO 0)"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("      );"));
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("END COMPONENT;"));
      WRITELINE(sp3_fout, line_out);
      WRITELINE(sp3_fout, line_out);
      FOR index IN 63 DOWNTO 0 LOOP
        WRITE(line_out, STRING'("ATTRIBUTE INIT_"));
        WRITE(line_out, vec2char(STD_LOGIC_VECTOR(CONV_SIGNED(index / 16, 4))));
        WRITE(line_out, vec2char(STD_LOGIC_VECTOR(CONV_SIGNED(index MOD 16, 4))));
        WRITE(line_out, STRING'(" : STRING;"));
        WRITELINE(sp3_fout, line_out);
      END LOOP;
      WRITELINE(sp3_fout, line_out);
      FOR index2 IN (16 / (data_msb + 1)) - 1 DOWNTO 0 LOOP
        FOR index IN 16383 DOWNTO 0 LOOP
          IF ((index+1) MOD 256 = 0) THEN
            WRITE(line_out, STRING'("ATTRIBUTE INIT_"));
            WRITE(line_out, vec2char(STD_LOGIC_VECTOR(CONV_SIGNED((index / 256) / 16, 4))));
            WRITE(line_out, vec2char(STD_LOGIC_VECTOR(CONV_SIGNED((index / 256) MOD 16, 4))));
            WRITE(line_out, STRING'(" OF U"));
            WRITE(line_out, index2);
            WRITE(line_out, STRING'(" : LABEL IS """));
          END IF;
          vec(index MOD 4) := inst_rec.list(index / (data_msb + 1)).inst(index MOD (data_msb + 1) + index2 * (data_msb + 1));
          IF (index MOD 4 = 0) THEN
            WRITE(line_out, vec2char(vec));
          END IF;
          IF (index MOD 256 = 0) THEN
            WRITE(line_out, STRING'(""";"));
            WRITELINE(sp3_fout, line_out);
          END IF;
        END LOOP;
        WRITELINE(sp3_fout, line_out);
      END LOOP;
      WRITE(line_out, STRING'("SIGNAL rst : STD_LOGIC;"));
      WRITELINE(sp3_fout, line_out);
      IF (one_more /= 0) THEN
        WRITE(line_out, STRING'("SIGNAL dip : STD_LOGIC_VECTOR("));
        WRITE(line_out, one_more-1);
        WRITE(line_out, STRING'(" DOWNTO 0);"));
        WRITELINE(sp3_fout, line_out);
      END IF;
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("BEGIN"));
      WRITELINE(sp3_fout, line_out);
      WRITELINE(sp3_fout, line_out);
      WRITE(line_out, STRING'("  rst <= '0';"));
      WRITELINE(sp3_fout, line_out);
      IF (one_more /= 0) THEN
        WRITE(line_out, STRING'("  dip <= """));
        FOR index IN 0 to one_more-1 LOOP
          WRITE(line_out, STRING'("0"));
        END LOOP;
        WRITE(line_out, STRING'(""";"));
        WRITELINE(sp3_fout, line_out);
      END IF;
      WRITELINE(sp3_fout, line_out);
      FOR index IN (16 / (data_msb + 1)) - 1 DOWNTO 0 LOOP
        WRITE(line_out, STRING'("  U"));
        WRITE(line_out, index);
        WRITE(line_out, STRING'(" : RAMB16_S"));
        WRITE(line_out, data_msb + 1 + one_more);
        WRITELINE(sp3_fout, line_out);
        WRITE(line_out, STRING'("    PORT MAP"));
        WRITELINE(sp3_fout, line_out);
        WRITE(line_out, STRING'("      (ssr  => rst,"));
        WRITELINE(sp3_fout, line_out);
        WRITE(line_out, STRING'("       clk  => clk,"));
        WRITELINE(sp3_fout, line_out);
        WRITE(line_out, STRING'("       en   => en,"));
        WRITELINE(sp3_fout, line_out);
        WRITE(line_out, STRING'("       we   => we,"));
        WRITELINE(sp3_fout, line_out);
        WRITE(line_out, STRING'("       addr => addr("));
        WRITE(line_out, addr_msb);
        WRITE(line_out, STRING'(" DOWNTO 0),"));
        WRITELINE(sp3_fout, line_out);
        IF (one_more /= 0) THEN
          WRITE(line_out, STRING'("       dip  => dip,"));
          WRITELINE(sp3_fout, line_out);
          WRITE(line_out, STRING'("       dop  => OPEN,"));
          WRITELINE(sp3_fout, line_out);
        END IF;
        WRITE(line_out, STRING'("       di   => din("));
        WRITE(line_out, (data_msb + 1)*(index+1)-1 );
        WRITE(line_out, STRING'(" DOWNTO "));
        WRITE(line_out, (data_msb + 1)*index);
        WRITE(line_out, STRING'("),"));
        WRITELINE(sp3_fout, line_out);
        WRITE(line_out, STRING'("       do   => dout("));
        WRITE(line_out, (data_msb + 1)*(index+1)-1 );
        WRITE(line_out, STRING'(" DOWNTO "));
        WRITE(line_out, (data_msb + 1)*index);
        WRITE(line_out, STRING'(")"));
        WRITELINE(sp3_fout, line_out);
        WRITE(line_out, STRING'("      );"));
        WRITELINE(sp3_fout, line_out);
        WRITELINE(sp3_fout, line_out);
      END LOOP;
      WRITE(line_out, STRING'("END rtl;"));
      WRITELINE(sp3_fout, line_out);

    ELSE

      WRITE(line_out, STRING'("NOT creating pre-initialised Block RAM VHDL file "));
      WRITE(line_out, source_path);
      WRITE(line_out, sp2_file);
      WRITE(line_out, STRING'(" for use as a "));
      WRITE(line_out, ram_size);
      WRITE(line_out, STRING'(" word Spartan 3 FPGA RAM (WebPack) - RAM TOO BIG."));
      WRITELINE(OUTPUT, line_out);
      WRITELINE(OUTPUT, line_out);

    END IF;

    WRITE(line_out, STRING'("Creating COE (RAM Content) file "));
    WRITE(line_out, coregen_path);
    WRITE(line_out, coe_file);
    WRITE(line_out, STRING'(" for use in creating a "));
    WRITE(line_out, ram_size);
    WRITE(line_out, STRING'(" word RAM with CoreGen."));
    WRITELINE(OUTPUT, line_out);
    WRITELINE(OUTPUT, line_out);

    --
    -- Write out binary to file and also copy binary into RAM model.
    --
    WRITE(line_out, STRING'("memory_initialization_radix=2;"));
    WRITELINE(coe_fout, line_out);
    WRITE(line_out, STRING'("memory_initialization_vector="));
    WRITELINE(coe_fout, line_out);
    FOR index IN 0 TO ram_size-1 LOOP
      memory_tmp(index) <= inst_rec.list(index).inst;
      WRITE(line_out, TO_BITVECTOR(inst_rec.list(index).inst));
      IF (index = ram_size - 1) THEN
        WRITE(line_out, STRING'(";"));
      ELSE
        WRITE(line_out, STRING'(","));
      END IF;
      WRITELINE(coe_fout, line_out);
    END LOOP;

    WRITE(line_out, STRING'("Creating XCO (Instruction) file "));
    WRITE(line_out, coregen_path);
    WRITE(line_out, xco_file);
    WRITE(line_out, STRING'(" for use in creating a "));
    WRITE(line_out, ram_size);
    WRITE(line_out, STRING'(" word RAM with CoreGen."));
    WRITELINE(OUTPUT, line_out);
    WRITELINE(OUTPUT, line_out);

    --
    -- Create Xilinx XCO file for building CoreGen RAM.
    --
    WRITE(line_out, STRING'("SET BusFormat = BusFormatAngleBracket"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("SET SimulationOutputProducts = VHDL"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("SET XilinxFamily = Spartan2"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("SET OutputOption = DesignFlow"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("SET DesignFlow = VHDL"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("SET FlowVendor = Synopsys"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("SET FormalVerification = None"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("SELECT Single_Port_Block_Memory Spartan2 Xilinx,_Inc. 4.0"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET width = 16"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET port_configuration = Read_And_Write"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET init_pin = false"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET enable_pin = true"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET coefficient_file = "));
    WRITE(line_out, coe_file);
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET depth = "));
    WRITE(line_out, ram_size);
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET has_limit_data_pitch = false"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET handshaking_pins = false"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET write_mode = Read_After_Write"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET additional_output_pipe_stages = 0"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET global_init_value = 0"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET component_name = memory"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET init_value = 0"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET register_inputs = false"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET load_init_file = true"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("CSET limit_data_pitch = 8"));
    WRITELINE(xco_fout, line_out);
    WRITE(line_out, STRING'("GENERATE"));
    WRITELINE(xco_fout, line_out);
    WRITELINE(xco_fout, line_out);

    WAIT;
  END PROCESS do_read;

  --
  -- RAM Model.
  --
  do_access : PROCESS(clk)
  BEGIN
    IF RISING_EDGE(clk) THEN
      IF (NOW < 100 ns) THEN
        memory <= memory_tmp;
      ELSIF (en = '1') THEN
        IF (we = '1') THEN
          memory(CONV_INTEGER(addr)) <= din;
          dout_i <= din;
        ELSE
          dout_i <= memory(CONV_INTEGER(addr));
        END If;
      END IF;
    END IF;
  END PROCESS do_access;

  dout <= dout_i;

END behave;
