---====================== Start Software License ======================---
--==                                                                  ==--
--== This license governs the use of this software, and your use of   ==--
--== this software constitutes acceptance of this license. Agreement  ==--
--== with all points is required to use this software.                ==--
--==                                                                  ==--
--== 1. You may use this software freely for personal use.            ==--
--==                                                                  ==--
--== 2. You may use this software freely to determine feasibility for ==--
--==    commercial use.                                               ==--
--==                                                                  ==--
--== 3. You may use this software for commercial use if the author    ==--
--==    has given you written consent.                                ==--
--==                                                                  ==--
--== 4. You may modify this software provided you do not remove the   ==--
--==    license and copyright notice.                                 ==--
--==                                                                  ==--
--== 5. You may distribute this software and derivative works to      ==--
--==    personal friends and work colleagues only.                    ==--
--==                                                                  ==--
--== 6. You agree that this software comes as-is and with no        ==--
--==    warranty whatsoever, either expressed or implied, including,  ==--
--==    but not limited to, warranties of merchantability or fitness  ==--
--==    for a particular purpose.                                     ==--
--==                                                                  ==--
--== 7. You agree that the author will not be liable for any damages  ==--
--==    relating from the use of this software, including direct,     ==--
--==    indirect, consequential or incidental. This software is used  ==--
--==    entirely at your own risk and should it prove defective, you  ==--
--==    will assume full responsibility for all costs associated with ==--
--==    servicing, repair or correction.                              ==--
--==                                                                  ==--
--== Your rights under this license are terminated immediately if you ==--
--== breach it in any way.                                            ==--
--==                                                                  ==--
---======================= End Software License =======================---


---====================== Start Copyright Notice ======================---
--==                                                                  ==--
--== Filename ..... raptor16_lcd_top.vhd                              ==--
--== Download ..... http://www.spacewire.co.uk                        ==--
--== Author ....... Steve Haywood (steve.haywood@ukonline.co.uk)      ==--
--== Copyright .... Copyright (c) 2004 Steve Haywood                  ==--
--== Project ...... Raptor-16                                         ==--
--== Version ...... 1.00                                              ==--
--== Conception ... 24 August 2004                                    ==--
--== Modified ..... N/A                                               ==--
--==                                                                  ==--
---======================= End Copyright Notice =======================---


---========================= Start Description ========================---
--==                                                                  ==--
--== Raptor-16 top level wrapper for LCD demonstration code.          ==--
--==                                                                  ==--
---========================== End Description =========================---

LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;

ENTITY raptor16_lcd_top IS
  PORT(
       --== General Interface (Async Rst, 50MHz Clock) ==--

       clk   : IN  STD_LOGIC;
       rst_n : IN  STD_LOGIC;

       --== LCD Interface (4-Bit) ==--

       lcd_rs   : OUT STD_LOGIC;
       lcd_rw   : OUT STD_LOGIC;
       lcd_en   : OUT STD_LOGIC;
       lcd_data : OUT STD_LOGIC_VECTOR(7 DOWNTO 4)
      );
END raptor16_lcd_top;


ARCHITECTURE rtl OF raptor16_lcd_top IS

---==========================---
--== Component Declarations ==--
---==========================---

COMPONENT raptor16
  PORT(--== General Interface (Sync Rst, 50MHz Clock) ==--
       clk         : IN  STD_LOGIC;
       rst         : IN  STD_LOGIC;
       --== User Port 0 Interface ==--
       port0_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port0_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port0_read  : OUT STD_LOGIC;
       port0_write : OUT STD_LOGIC;
       --== User Port 1 Interface ==--
       port1_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port1_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port1_read  : OUT STD_LOGIC;
       port1_write : OUT STD_LOGIC;
       --== User Port 2 Interface ==--
       port2_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port2_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port2_read  : OUT STD_LOGIC;
       port2_write : OUT STD_LOGIC;
       --== User Port 3 Interface ==--
       port3_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port3_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port3_read  : OUT STD_LOGIC;
       port3_write : OUT STD_LOGIC;
       --== User Port 4 Interface ==--
       port4_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port4_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port4_read  : OUT STD_LOGIC;
       port4_write : OUT STD_LOGIC;
       --== Interrupt Priority 1 (Lowest) Interface ==--
       int_req1    : IN  STD_LOGIC;
       int_ack1    : OUT STD_LOGIC;
       --== Interrupt Priority 2 Interface ==--
       int_req2    : IN  STD_LOGIC;
       int_ack2    : OUT STD_LOGIC;
       --== Interrupt Priority 3 Interface ==--
       int_req3    : IN  STD_LOGIC;
       int_ack3    : OUT STD_LOGIC;
       --== Interrupt Priority 4 Interface ==--
       int_req4    : IN  STD_LOGIC;
       int_ack4    : OUT STD_LOGIC;
       --== Interrupt Priority 5 Interface ==--
       int_req5    : IN  STD_LOGIC;
       int_ack5    : OUT STD_LOGIC;
       --== Interrupt Priority 6 Interface ==--
       int_req6    : IN  STD_LOGIC;
       int_ack6    : OUT STD_LOGIC;
       --== Interrupt Priority 7 (Highest) Interface ==--
       int_req7    : IN  STD_LOGIC;
       int_ack7    : OUT STD_LOGIC;
       --== Static RAM Interface ==--
       ram_en      : OUT STD_LOGIC;
       ram_we      : OUT STD_LOGIC;
       ram_addr    : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       ram_dout    : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       ram_din     : OUT STD_LOGIC_VECTOR(15 DOWNTO 0)
      );
END COMPONENT;

COMPONENT memory
  PORT(clk  : IN  STD_LOGIC;
       en   : IN  STD_LOGIC;
       we   : IN  STD_LOGIC;
       addr : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       din  : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       dout : OUT STD_LOGIC_VECTOR(15 DOWNTO 0)
      );
END COMPONENT;

---=======================---
--== Signal Declarations ==--
---=======================---

SIGNAL rst        : STD_LOGIC := '0';
SIGNAL rst_ff     : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL int_req    : STD_LOGIC_VECTOR(7 DOWNTO 1);
SIGNAL int_ack    : STD_LOGIC_VECTOR(7 DOWNTO 1);
SIGNAL port0_data : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL port1_data : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL port2_data : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL port3_data : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL port4_data : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL ram_en     : STD_LOGIC;
SIGNAL ram_we     : STD_LOGIC;
SIGNAL ram_addr   : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL ram_din    : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL ram_dout   : STD_LOGIC_VECTOR(15 DOWNTO 0);

BEGIN

  ---====================================================---
  --== Board specific logic - Syncronize & invert Reset ==--
  ---====================================================---

  PROCESS(clk, rst_n)
  BEGIN
    IF (rst_n = '0') THEN
      rst_ff <= "110";
    ELSIF RISING_EDGE(clk) THEN
      rst_ff <= rst_ff(1 DOWNTO 0) & '0';
    END IF;
  END PROCESS;

  PROCESS(clk)
  BEGIN
    IF RISING_EDGE(clk) THEN
      rst <= rst_ff(2);
    END IF;
  END PROCESS;

  ---==========================---
  --== Connect up LCD Signals ==--
  ---==========================---

  lcd_rs   <= port0_data(0);
  lcd_en   <= port0_data(1);
  lcd_rw   <= port0_data(2);
  lcd_data <= port0_data(6 DOWNTO 3);

  ---==============================================---
  --== Tie interrupt requests low (no interrupts) ==--
  ---==============================================---

  int_req <= (OTHERS => '0');

  ---====================================---
  --== Raptor-16 Embeded Microprocessor ==--
  ---====================================---

  U0 : raptor16
    PORT MAP
      (--== General Interface (Sync Rst, 50MHz Clock) ==--
       clk         => clk,
       rst         => rst,
       --== User Port 0 Interface ==--
       port0_din   => port0_data,
       port0_dout  => port0_data,
       port0_read  => OPEN,
       port0_write => OPEN,
       --== User Port 1 Interface ==--
       port1_din   => port1_data,
       port1_dout  => port1_data,
       port1_read  => OPEN,
       port1_write => OPEN,
       --== User Port 2 Interface ==--
       port2_din   => port2_data,
       port2_dout  => port2_data,
       port2_read  => OPEN,
       port2_write => OPEN,
       --== User Port 3 Interface ==--
       port3_din   => port3_data,
       port3_dout  => port3_data,
       port3_read  => OPEN,
       port3_write => OPEN,
       --== User Port 4 Interface ==--
       port4_din   => port4_data,
       port4_dout  => port4_data,
       port4_read  => OPEN,
       port4_write => OPEN,
       --== Interrupt Priority 1 (Lowest) Interface ==--
       int_req1    => int_req(1),
       int_ack1    => int_ack(1),
       --== Interrupt Priority 2 Interface ==--
       int_req2    => int_req(2),
       int_ack2    => int_ack(2),
       --== Interrupt Priority 3 Interface ==--
       int_req3    => int_req(3),
       int_ack3    => int_ack(3),
       --== Interrupt Priority 4 Interface ==--
       int_req4    => int_req(4),
       int_ack4    => int_ack(4),
       --== Interrupt Priority 5 Interface ==--
       int_req5    => int_req(5),
       int_ack5    => int_ack(5),
       --== Interrupt Priority 6 Interface ==--
       int_req6    => int_req(6),
       int_ack6    => int_ack(6),
       --== Interrupt Priority 7 (Highest) Interface ==--
       int_req7    => int_req(7),
       int_ack7    => int_ack(7),
       --== Static RAM Interface ==--
       ram_en      => ram_en,
       ram_we      => ram_we,
       ram_addr    => ram_addr,
       ram_dout    => ram_dout,
       ram_din     => ram_din
      );

  ---======================---
  --== Embeded Static RAM ==--
  ---======================---

  U1 : memory
    PORT MAP
      (clk  => clk,
       en   => ram_en,
       we   => ram_we,
       addr => ram_addr,
       din  => ram_din,
       dout => ram_dout
      );

END rtl;
